# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
# Authors: Francesco Bartoli <xbartolone@gmail.com>
# Authors: Ricardo Garcia Silva <ricardo.garcia.silva@geobeyond.it>
#
# Copyright (c) 2026 Tom Kralidis
# Copyright (c) 2025 Francesco Bartoli
# Copyright (c) 2023 Ricardo Garcia Silva
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from copy import deepcopy
import io
import json
import logging
import os
from pathlib import Path
from typing import Union

import click
from jsonschema import validate as jsonschema_validate
import yaml

from pygeoapi import l10n
from pygeoapi.api import all_apis
from pygeoapi.models.openapi import OAPIFormat
from pygeoapi.util import (filter_dict_by_key_value, to_json, yaml_load,
                           get_api_rules, get_base_url, SCHEMASDIR)

LOGGER = logging.getLogger(__name__)

OPENAPI_YAML = {
    'cql2': 'https://schemas.opengis.net/cql2/1.0/cql2.json',
    'oapif-1': 'https://schemas.opengis.net/ogcapi/features/part1/1.0/openapi/ogcapi-features-1.yaml',  # noqa
    'oapif-2': 'https://schemas.opengis.net/ogcapi/features/part2/1.0/openapi/ogcapi-features-2.yaml', # noqa
    'oapip': 'https://schemas.opengis.net/ogcapi/processes/part1/1.0/openapi',
    'oacov': 'https://raw.githubusercontent.com/opengeospatial/ogcapi-coverages/refs/heads/master/standard/openapi/ogcapi-coverages-1.yaml',  # noqa
    'oamaps': 'https://schemas.opengis.net/ogcapi/maps/part1/1.0/openapi/ogcapi-maps-1.yaml',  # noqa
    'oapir': 'https://raw.githubusercontent.com/opengeospatial/ogcapi-records/master/core/openapi',  # noqa
    'oaedr': 'https://schemas.opengis.net/ogcapi/edr/1.0/openapi', # noqa
    'oapit': 'https://schemas.opengis.net/ogcapi/tiles/part1/1.0/openapi/ogcapi-tiles-1.yaml',  # noqa
    'pygeoapi': 'https://raw.githubusercontent.com/geopython/pygeoapi/master/pygeoapi/schemas/config/pygeoapi-config-0.x.yml'  # noqa
}


def get_ogc_schemas_location(server_config: dict) -> str:
    """
    Determine OGC schemas location

    :param server_config: `dict` of server configuration

    :returns: `str` of OGC schemas location
    """

    osl = server_config.get('ogc_schemas_location')

    value = 'https://schemas.opengis.net'

    if osl is not None:
        if osl.startswith('http'):
            value = osl
        elif osl.startswith('/'):
            base_url = get_base_url({'server': server_config})
            value = f'{base_url}/schemas'

    return value


# TODO: remove this function once OGC API - Processing is final
def gen_media_type_object(media_type: str, api_type: str, path: str) -> dict:
    """
    Generates an OpenAPI Media Type Object

    :param media_type: MIME type
    :param api_type: OGC API type
    :param path: local path of OGC API parameter or schema definition

    :returns: `dict` of media type object
    """

    ref = f'{OPENAPI_YAML[api_type]}/{path}'

    content = {
        media_type: {
            'schema': {
                '$ref': ref
            }
        }
    }

    return content


# TODO: remove this function once OGC API - Processing is final
def gen_response_object(description: str, media_type: str,
                        api_type: str, path: str) -> dict:
    """
    Generates an OpenAPI Response Object

    :param description: text description of response
    :param media_type: MIME type
    :param api_type: OGC API type

    :returns: `dict` of response object
    """

    response = {
        'description': description,
        'content': gen_media_type_object(media_type, api_type, path)
    }

    return response


def gen_contact(cfg: dict) -> dict:
    """
    Generates an OpenAPI contact object with OGC extensions
    based on OGC API - Records contact

    :param cfg: `dict` of configuration

    :returns: `dict` of OpenAPI contact object
    """

    has_addresses = False
    has_phones = False

    contact = {
        'name': cfg['metadata']['provider']['name']
    }

    for key in ['url', 'email']:
        if key in cfg['metadata']['provider']:
            contact[key] = cfg['metadata']['provider'][key]

    contact['x-ogc-serviceContact'] = {
        'name': cfg['metadata']['contact']['name'],
        'addresses': []
    }

    if 'position' in cfg['metadata']['contact']:
        contact['x-ogc-serviceContact']['position'] = cfg['metadata']['contact']['position']  # noqa

    if any(address in ['address', 'city', 'stateorprovince', 'postalcode', 'country'] for address in cfg['metadata']['contact']):  # noqa
        has_addresses = True

    if has_addresses:
        address = {}
        if 'address' in cfg['metadata']['contact']:
            address['deliveryPoint'] = [cfg['metadata']['contact']['address']]

        if 'city' in cfg['metadata']['contact']:
            address['city'] = cfg['metadata']['contact']['city']

        if 'stateorprovince' in cfg['metadata']['contact']:
            address['administrativeArea'] = cfg['metadata']['contact']['stateorprovince']  # noqa

        if 'postalCode' in cfg['metadata']['contact']:
            address['administrativeArea'] = cfg['metadata']['contact']['postalCode']  # noqa

        if 'country' in cfg['metadata']['contact']:
            address['administrativeArea'] = cfg['metadata']['contact']['country']  # noqa

        contact['x-ogc-serviceContact']['addresses'].append(address)

    if any(phone in ['phone', 'fax'] for phone in cfg['metadata']['contact']):
        has_phones = True
        contact['x-ogc-serviceContact']['phones'] = []

    if has_phones:
        if 'phone' in cfg['metadata']['contact']:
            contact['x-ogc-serviceContact']['phones'].append({
                'type': 'main', 'value': cfg['metadata']['contact']['phone']
            })

        if 'fax' in cfg['metadata']['contact']:
            contact['x-ogc-serviceContact']['phones'].append({
                'type': 'fax', 'value': cfg['metadata']['contact']['fax']
            })

    if 'email' in cfg['metadata']['contact']:
        contact['x-ogc-serviceContact']['emails'] = [{
            'value': cfg['metadata']['contact']['email']
        }]

    if 'url' in cfg['metadata']['contact']:
        contact['x-ogc-serviceContact']['links'] = [{
            'type': 'text/html',
            'href': cfg['metadata']['contact']['url']
        }]

    if 'instructions' in cfg['metadata']['contact']:
        contact['x-ogc-serviceContact']['contactInstructions'] = cfg['metadata']['contact']['instructions']  # noqa

    if 'hours' in cfg['metadata']['contact']:
        contact['x-ogc-serviceContact']['hoursOfService'] = cfg['metadata']['contact']['hours']  # noqa

    if 'role' in cfg['metadata']['contact']:
        contact['x-ogc-serviceContact']['hoursOfService'] = cfg['metadata']['contact']['role']  # noqa

    return contact


def get_oas_30(cfg: dict, fail_on_invalid_collection: bool = True) -> dict:
    """
    Generates an OpenAPI 3.0 Document

    :param cfg: configuration object
    :param fail_on_invalid_collection: `bool` of whether to fail on an invalid
                                       collection

    :returns: dict of OpenAPI definition
    """

    paths = {}

    # TODO: make openapi multilingual (default language only for now)
    locale_ = l10n.get_locales(cfg)[0]

    api_rules = get_api_rules(cfg)

    osl = get_ogc_schemas_location(cfg['server'])
    OPENAPI_YAML['oapif-1'] = os.path.join(osl, 'ogcapi/features/part1/1.0/openapi/ogcapi-features-1.yaml')  # noqa
    OPENAPI_YAML['oapif-2'] = os.path.join(osl, 'ogcapi/features/part2/1.0/openapi/ogcapi-features-2.yaml') # noqa

    LOGGER.debug('setting up server info')
    oas = {
        'openapi': '3.0.2',
        'tags': []
    }
    info = {
        'title': l10n.translate(cfg['metadata']['identification']['title'], locale_),  # noqa
        'description': l10n.translate(cfg['metadata']['identification']['description'], locale_),  # noqa
        'x-keywords': l10n.translate(cfg['metadata']['identification']['keywords'], locale_),  # noqa
        'termsOfService':
            cfg['metadata']['identification']['terms_of_service'],
        'contact': gen_contact(cfg),
        'license': {
            'name': cfg['metadata']['license']['name'],
            'url': cfg['metadata']['license']['url']
        },
        'version': api_rules.api_version
    }
    oas['info'] = info

    oas['servers'] = [{
        'url': get_base_url(cfg),
        'description': l10n.translate(cfg['metadata']['identification']['description'], locale_)  # noqa
    }]

    paths['/'] = {
        'get': {
            'summary': 'Landing page',
            'description': 'Landing page',
            'tags': ['server'],
            'operationId': 'getLandingPage',
            'parameters': [
                {'$ref': '#/components/parameters/f'},
                {'$ref': '#/components/parameters/lang'}
            ],
            'responses': {
                '200': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/LandingPage"},  # noqa
                '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"}  # noqa
            }
        }
    }

    paths['/openapi'] = {
        'get': {
            'summary': 'This document',
            'description': 'This document',
            'tags': ['server'],
            'operationId': 'getOpenapi',
            'parameters': [
                {'$ref': '#/components/parameters/f'},
                {'$ref': '#/components/parameters/lang'},
                {
                    'name': 'ui',
                    'in': 'query',
                    'description': 'UI to render the OpenAPI document',
                    'required': False,
                    'schema': {
                        'type': 'string',
                        'enum': ['swagger', 'redoc'],
                        'default': 'swagger'
                    },
                    'style': 'form',
                    'explode': False
                },
            ],
            'responses': {
                '200': {'$ref': '#/components/responses/200'},
                '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                'default': {'$ref': '#/components/responses/default'}
            }
        }
    }

    paths['/conformance'] = {
        'get': {
            'summary': 'API conformance definition',
            'description': 'API conformance definition',
            'tags': ['server'],
            'operationId': 'getConformanceDeclaration',
            'parameters': [
                {'$ref': '#/components/parameters/f'},
                {'$ref': '#/components/parameters/lang'}
            ],
            'responses': {
                '200': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/LandingPage"},  # noqa
                '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"}  # noqa
            }
        }
    }

    paths['/collections'] = {
        'get': {
            'summary': 'Collections',
            'description': 'Collections',
            'tags': ['server'],
            'operationId': 'getCollections',
            'parameters': [
                {'$ref': '#/components/parameters/f'},
                {'$ref': '#/components/parameters/lang'}
            ],
            'responses': {
                '200': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/LandingPage"},  # noqa
                '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"}  # noqa
            }
        }
    }

    oas['tags'].append({
            'name': 'server',
            'description': l10n.translate(cfg['metadata']['identification']['description'], locale_),  # noqa
            'externalDocs': {
                'description': 'information',
                'url': cfg['metadata']['identification']['url']}
        }
    )

    oas['components'] = {
        'responses': {
            '200': {
                'description': 'successful operation'
            },
            '204': {
                'description': 'no content'
            },
            'default': {
                'description': 'Unexpected error',
                'content': gen_media_type_object('application/json', 'oapip', 'schemas/exception.yaml')  # noqa
            },
            'Queryables': {
                'description': 'successful queryables operation',
                'content': {
                    'application/json': {
                        'schema': {'$ref': '#/components/schemas/queryables'}
                    }
                }
            }
        },
        'parameters': get_oas_30_parameters(cfg=cfg, locale_=locale_),
        'schemas': {
            # TODO: change this schema once OGC will definitively publish it
            'queryable': {
                'type': 'object',
                'required': [
                    'queryable',
                    'type'
                ],
                'properties': {
                    'queryable': {
                        'description': 'the token that may be used in a CQL predicate', # noqa
                        'type': 'string'
                    },
                    'title': {
                        'description': 'a human readable title for the queryable', # noqa
                        'type': 'string'
                    },
                    'description': {
                        'description': 'a human-readable narrative describing the queryable', # noqa
                        'type': 'string'
                    },
                    'language': {
                        'description': 'the language used for the title and description', # noqa
                        'type': 'string',
                        'default': 'en'
                    },
                    'type': {
                        'description': 'the data type of the queryable', # noqa
                        'type': 'string'
                    },
                    'type-ref': {
                        'description': 'a reference to the formal definition of the type', # noqa
                        'type': 'string',
                        'format': 'url'
                    }
                }
            },
            'queryables': {
                'type': 'object',
                'required': [
                    'queryables'
                ],
                'properties': {
                    'queryables': {
                        'type': 'array',
                        'items': {'$ref': '#/components/schemas/queryable'}
                    }
                }
            }
        }
    }

    items_f = deepcopy(oas['components']['parameters']['f'])
    items_f['schema']['enum'].append('csv')

    LOGGER.debug('setting up datasets')

    for k, v in get_visible_collections(cfg).items():
        name = l10n.translate(k, locale_)
        title = l10n.translate(v['title'], locale_)
        desc = l10n.translate(v['description'], locale_)
        collection_name_path = f'/collections/{k}'
        tag = {
            'name': name,
            'description': desc,
            'externalDocs': {}
        }
        for link in l10n.translate(v.get('links', []), locale_):
            if link['type'] == 'information':
                tag['externalDocs']['description'] = link['type']
                tag['externalDocs']['url'] = link['url']
                break
        if len(tag['externalDocs']) == 0:
            del tag['externalDocs']

        oas['tags'].append(tag)

        paths[collection_name_path] = {
            'get': {
                'summary': f'Get {title} metadata',
                'description': desc,
                'tags': [name],
                'operationId': f'describe{name.capitalize()}Collection',
                'parameters': [
                    {'$ref': '#/components/parameters/f'},
                    {'$ref': '#/components/parameters/lang'}
                ],
                'responses': {
                    '200': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/Collection"},  # noqa
                    '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                    '404': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/NotFound"},  # noqa
                    '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"}  # noqa
                }
            }
        }

        oas['components']['responses'].update({
                'Tiles': {
                    'description': 'Retrieves the tiles description for this collection', # noqa
                    'content': {
                        'application/json': {
                            'schema': {
                                '$ref': '#/components/schemas/tiles'
                            }
                        }
                    }
                }
            }
        )

        oas['components']['schemas'].update({
                'tilematrixsetlink': {
                    'type': 'object',
                    'required': ['tileMatrixSet'],
                    'properties': {
                        'tileMatrixSet': {
                            'type': 'string'
                        },
                        'tileMatrixSetURI': {
                            'type': 'string'
                        }
                    }
                },
                'tiles': {
                    'type': 'object',
                    'required': [
                        'tileMatrixSetLinks',
                        'links'
                    ],
                    'properties': {
                        'tileMatrixSetLinks': {
                            'type': 'array',
                            'items': {
                                '$ref': '#/components/schemas/tilematrixsetlink' # noqa
                            }
                        },
                        'links': {
                            'type': 'array',
                            'items': {'$ref': f"{OPENAPI_YAML['oapit']}#/components/schemas/link"}  # noqa
                        }
                    }
                }
            }
        )

    oas['paths'] = paths

    for api_name, api_module in all_apis().items():
        LOGGER.debug(f'Adding OpenAPI definitions for {api_name}')

        try:
            sub_tags, sub_paths = api_module.get_oas_30(cfg, locale_)

            if not sub_tags and not sub_paths:
                LOGGER.debug('Empty content from {api_name}; skipping')
                continue

            oas['paths'].update(sub_paths['paths'])
            oas['tags'].extend(sub_tags)
        except Exception as err:
            if fail_on_invalid_collection:
                raise
            else:
                LOGGER.warning(f'Resource not added to OpenAPI: {err}')

    if cfg['server'].get('admin', False):
        schema_dict = get_config_schema()
        oas['definitions'] = schema_dict['definitions']
        LOGGER.debug('Adding admin endpoints')
        oas['paths'].update(get_admin(cfg))

    return oas


def get_oas_30_parameters(cfg: dict, locale_: str):
    server_locales = l10n.get_locales(cfg)

    oas_30_parameters = {
        'f': {
            'name': 'f',
            'in': 'query',
            'description': 'The optional f parameter indicates the output format which the server shall provide as part of the response document.  The default format is GeoJSON.',  # noqa
            'required': False,
            'schema': {
                'type': 'string',
                'enum': ['json', 'html', 'jsonld'],
                'default': 'json'
            },
            'style': 'form',
            'explode': False
        },
        'lang': {
            'name': 'lang',
            'in': 'query',
            'description': 'The optional lang parameter instructs the server return a response in a certain language, if supported.  If the language is not among the available values, the Accept-Language header language will be used if it is supported. If the header is missing, the default server language is used. Note that providers may only support a single language (or often no language at all), that can be different from the server language.  Language strings can be written in a complex (e.g. "fr-CA,fr;q=0.9,en-US;q=0.8,en;q=0.7"), simple (e.g. "de") or locale-like (e.g. "de-CH" or "fr_BE") fashion.',  # noqa
            'required': False,
            'schema': {
                'type': 'string',
                'enum': [l10n.locale2str(sl) for sl in server_locales],
                'default': l10n.locale2str(locale_)
            }
        },
        'skipGeometry': {
            'name': 'skipGeometry',
            'in': 'query',
            'description': 'This option can be used to skip response geometries for each feature.',  # noqa
            'required': False,
            'style': 'form',
            'explode': False,
            'schema': {
                'type': 'boolean',
                'default': False
            }
        },
        'crs': {
            'name': 'crs',
            'in': 'query',
            'description': 'Indicates the coordinate reference system for the results.',  # noqa
            'style': 'form',
            'required': False,
            'explode': False,
            'schema': {
                'format': 'uri',
                'type': 'string'
            }
        },
        'bbox': {
            'name': 'bbox',
            'in': 'query',
            'description': 'Only features that have a geometry that intersects the bounding box are selected.'  # noqa
                           'The bounding box is provided as four or six numbers, depending on whether the '  # noqa
                           'coordinate reference system includes a vertical axis (height or depth).',  # noqa
            'required': False,
            'style': 'form',
            'explode': False,
            'schema': {
                'type': 'array',
                'minItems': 4,
                'maxItems': 6,
                'items': {
                    'type': 'number'
                }
            }
        },
        'bbox-crs': {
            'name': 'bbox-crs',
            'in': 'query',
            'description': 'Indicates the coordinate reference system for the given bbox coordinates.',  # noqa
            'style': 'form',
            'required': False,
            'explode': False,
            'schema': {
                'format': 'uri',
                'type': 'string'
            }
        },
        # FIXME: This is not compatible with the bbox-crs definition in
        #        OGCAPI Features Part 2!
        #        We need to change the mapscript provider and
        #        get_collection_map() method in the API!
        #        So this is for de map-provider only.
        'bbox-crs-epsg': {
            'name': 'bbox-crs',
            'in': 'query',
            'description': 'Indicates the EPSG for the given bbox coordinates.',  # noqa
            'required': False,
            'style': 'form',
            'explode': False,
            'schema': {
                'type': 'integer',
                'default': 4326
            }
        },
        'offset': {
            'name': 'offset',
            'in': 'query',
            'description': 'The optional offset parameter indicates the index within the result set from which the server shall begin presenting results in the response document.  The first element has an index of 0 (default).',  # noqa
            'required': False,
            'schema': {
                'type': 'integer',
                'minimum': 0,
                'default': 0
            },
            'style': 'form',
            'explode': False
        },
        'properties': {
            'name': 'properties',
            'in': 'query',
            'description': 'The properties that should be included. The parameter value is a comma-separated list of property names.',  # noqa
            'required': False,
            'style': 'form',
            'explode': False,
            'schema': {
                'type': 'array',
                'items': {
                    'type': 'string'
                }
            }
        },
        'vendorSpecificParameters': {
            'name': 'vendorSpecificParameters',
            'in': 'query',
            'description': 'Additional "free-form" parameters that are not explicitly defined',  # noqa
            'schema': {
                'type': 'object',
                'additionalProperties': True
            },
            'style': 'form'
        },
        'resourceId': {
            'name': 'resourceId',
            'in': 'path',
            'description': 'Configuration resource identifier',
            'required': True,
            'schema': {
                'type': 'string'
             }
        }
    }
    if len(list(cfg['resources'].keys())) > 0:
        oas_30_parameters['resourceId']['schema']['default'] = list(
            cfg['resources'].keys()
        )[0]
    return oas_30_parameters


def get_visible_collections(cfg: dict) -> dict:
    collections = filter_dict_by_key_value(cfg['resources'],
                                           'type', 'collection')

    return {
        k: v
        for k, v in collections.items()
        if v.get('visibility', 'default') != 'hidden'
    }


def get_config_schema():
    schema_file = SCHEMASDIR / 'config' / 'pygeoapi-config-0.x.yml'

    with schema_file.open() as fh2:
        return yaml_load(fh2)


def get_admin(cfg: dict) -> dict:

    schema_dict = get_config_schema()

    paths = {}
    if cfg['resources']:
        res_eg_key = next(iter(cfg['resources']))
    else:
        res_eg_key = 'example'
    res_eg = {
        res_eg_key: cfg['resources'][res_eg_key]
    } if cfg['resources'] else {
        'example': {
            'type': 'collection',
            'title': 'Example',
            'description': 'Example',
            'keywords': ['example'],
            'links': [],
            'linked-data': {},
            'extents': {
                'spatial': {
                    'bbox': [-180, -90, 180, 90],
                    'crs': 'http://www.opengis.net/def/crs/OGC/1.3/CRS84'
                },
                'temporal': {
                    'begin': '2000-10-30T18:24:39Z',
                    'end': '2007-10-30T08:57:29Z',
                    'trs': 'http://www.opengis.net/def/uom/ISO-8601/0/Gregorian'  # noqa
                }
            }
        }
    }
    if 'extents' in res_eg[res_eg_key]:
        res_eg_eg_key = 'extents'
    elif 'type' in res_eg[res_eg_key]:
        res_eg_eg_key = 'type'

    res_eg[res_eg_key]['patch_example'] = {
        res_eg_eg_key: res_eg[res_eg_key][res_eg_eg_key]
    }

    paths['/admin/config'] = {
        'get': {
            'summary': 'Get admin configuration',
            'description': 'Get admin configuration',
            'tags': ['admin'],
            'operationId': 'getAdminConfig',
            'parameters': [
                {'$ref': '#/components/parameters/f'},
                {'$ref': '#/components/parameters/lang'}
            ],
            'responses': {
                '200': {
                    'description': 'Successful response',
                    'content': {
                        'application/json': {
                            'schema': schema_dict
                        }
                    }
                }
            }
        },
        'put': {
            'summary': 'Update admin configuration full',
            'description': 'Update admin configuration full',
            'tags': ['admin'],
            'operationId': 'putAdminConfig',
            'requestBody': {
                'description': 'Updates admin configuration',
                'content': {
                    'application/json': {
                        'example': cfg,
                        'schema': schema_dict
                    }
                },
                'required': True
            },
            'responses': {
                '204': {'$ref': '#/components/responses/204'},
                '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"}  # noqa
            }
        },
        'patch': {
            'summary': 'Partially update admin configuration',
            'description': 'Partially update admin configuration',
            'tags': ['admin'],
            'operationId': 'patchAdminConfig',
            'requestBody': {
                'description': 'Updates admin configuration',
                'content': {
                    'application/json': {
                        'example': {'metadata': cfg['metadata']},
                        'schema': schema_dict
                    }
                },
                'required': True
            },
            'responses': {
                '204': {'$ref': '#/components/responses/204'},
                '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"}  # noqa
            }
        }
    }
    paths['/admin/config/resources'] = {
        'get': {
            'summary': 'Get admin configuration resources',
            'description': 'Get admin configuration resources',
            'tags': ['admin'],
            'operationId': 'getAdminConfigResources',
            'parameters': [
                {'$ref': '#/components/parameters/f'},
                {'$ref': '#/components/parameters/lang'}
            ],
            'responses': {
                '200': {
                    'description': 'Successful response',
                    'content': {
                        'application/json': {
                            'schema': schema_dict['properties']['resources']['patternProperties']['^.*$']  # noqa
                        }
                    }
                }
            }
        },
        'post': {
            'summary': 'Create admin configuration resource',
            'description': 'Create admin configuration resource',
            'tags': ['admin'],
            'operationId': 'postAdminConfigResource',
            'requestBody': {
                'description': 'Adds resource to configuration',
                'content': {
                    'application/json': {
                        'example': {'new-collection': cfg['resources'][res_eg_key] if cfg['resources'] else res_eg['example'] }, # noqa
                        'schema': schema_dict['properties']['resources']['patternProperties']['^.*$']  # noqa
                    }
                },
                'required': True
            },
            'responses': {
                '201': {'description': 'Successful creation'},
                '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"}  # noqa
            }
        },
    }
    paths['/admin/config/resources/{resourceId}'] = {
        'get': {
            'summary': 'Get admin configuration resource',
            'description': 'Get admin configuration resource',
            'tags': ['admin'],
            'operationId': 'getAdminConfigResource',
            'parameters': [
                {'$ref': '#/components/parameters/resourceId'},
                {'$ref': '#/components/parameters/f'},
                {'$ref': '#/components/parameters/lang'}
            ],
            'responses': {
                '200': {
                    'description': 'Successful response',
                    'content': {
                        'application/json': {
                            'schema': schema_dict['properties']['resources']['patternProperties']['^.*$']  # noqa
                        }
                    }
                }
            }
        },
        'put': {
            'summary': 'Update admin configuration resource',
            'description': 'Update admin configuration resource',
            'tags': ['admin'],
            'operationId': 'putAdminConfigResource',
            'parameters': [
                {'$ref': '#/components/parameters/resourceId'},
            ],
            'requestBody': {
                'description': 'Updates admin configuration resource',
                'content': {
                    'application/json': {
                        'example': res_eg[res_eg_key],
                        'schema': schema_dict['properties']['resources']['patternProperties']['^.*$']  # noqa
                    }
                },
                'required': True
            },
            'responses': {
                '204': {'$ref': '#/components/responses/204'},
                '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"}  # noqa
            }
        },
        'patch': {
            'summary': 'Partially update admin configuration resource',
            'description': 'Partially update admin configuration resource',
            'tags': ['admin'],
            'operationId': 'patchAdminConfigResource',
            'parameters': [
                {'$ref': '#/components/parameters/resourceId'},
            ],
            'requestBody': {
                'description': 'Updates admin configuration resource',
                'content': {
                    'application/json': {
                        'example': res_eg[res_eg_key]['patch_example'],
                        'schema': schema_dict['properties']['resources']['patternProperties']['^.*$']  # noqa
                    }
                },
                'required': True
            },
            'responses': {
                '204': {'$ref': '#/components/responses/204'},
                '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"}  # noqa
            }
        },
        'delete': {
            'summary': 'Delete admin configuration resource',
            'description': 'Delete admin configuration resource',
            'tags': ['admin'],
            'operationId': 'deleteAdminConfigResource',
            'parameters': [
                {'$ref': '#/components/parameters/resourceId'},
            ],
            'responses': {
                '204': {'$ref': '#/components/responses/204'},
                '404': {'$ref': f"{OPENAPI_YAML['oapip']}/responses/NotFound.yaml"},  # noqa
                'default': {'$ref': '#/components/responses/default'}  # noqa
            }
        }
    }

    return paths


def get_oas(cfg: dict, fail_on_invalid_collection: bool = True,
            version='3.0') -> dict:
    """
    Stub to generate OpenAPI Document

    :param cfg: `dict` configuration
    :param fail_on_invalid_collection: `bool` of whether to fail on an
                                       invalid collection
    :param version: version of OpenAPI (default 3.0)

    :returns: `dict` of OpenAPI definition
    """

    if version == '3.0':
        return get_oas_30(
            cfg, fail_on_invalid_collection=fail_on_invalid_collection)
    else:
        raise RuntimeError('OpenAPI version not supported')


def validate_openapi_document(instance_dict: dict) -> bool:
    """
    Validate an OpenAPI document against the OpenAPI schema

    :param instance_dict: dict of OpenAPI instance

    :returns: `bool` of validation
    """

    schema_file = SCHEMASDIR / 'openapi' / 'openapi-3.0.x.json'

    LOGGER.debug(f'Validating against {schema_file}')
    with schema_file.open() as fh2:
        schema_dict = json.load(fh2)
        jsonschema_validate(instance_dict, schema_dict)

        return True


def generate_openapi_document(cfg_file: Union[Path, io.TextIOWrapper],
                              output_format: OAPIFormat,
                              fail_on_invalid_collection: bool = True) -> str:
    """
    Generate an OpenAPI document from the configuration file

    :param cfg_file: configuration Path instance (`str` of filepath
                     or parsed `dict`)
    :param output_format: output format for OpenAPI document
    :param fail_on_invalid_collection: `bool` of whether to fail on an
                                       invalid collection

    :returns: `str` of the OpenAPI document in the output format requested
    """

    LOGGER.debug(f'Loading configuration {cfg_file}')

    if isinstance(cfg_file, Path):
        with cfg_file.open(mode="r") as cf:
            s = yaml_load(cf)
    else:
        s = yaml_load(cfg_file)

    pretty_print = s['server'].get('pretty_print', False)

    oas = get_oas(s, fail_on_invalid_collection=fail_on_invalid_collection)

    if output_format == 'yaml':
        content = yaml.safe_dump(oas, default_flow_style=False)
    else:
        content = to_json(oas, pretty=pretty_print)
    return content


def load_openapi_document() -> dict:
    """
    Open OpenAPI document from `PYGEOAPI_OPENAPI` environment variable

    :returns: `dict` of OpenAPI document
    """

    pygeoapi_openapi = os.environ.get('PYGEOAPI_OPENAPI')

    if pygeoapi_openapi is None:
        msg = 'PYGEOAPI_OPENAPI environment not set'
        LOGGER.error(msg)
        raise RuntimeError(msg)

    if not os.path.exists(pygeoapi_openapi):
        msg = (f'OpenAPI document {pygeoapi_openapi} does not exist.  '
               'Please generate before starting pygeoapi')
        LOGGER.error(msg)
        raise RuntimeError(msg)

    with open(pygeoapi_openapi, encoding='utf8') as ff:
        if pygeoapi_openapi.endswith(('.yaml', '.yml')):
            openapi_ = yaml_load(ff)
        else:  # JSON string, do not transform
            openapi_ = ff.read()

    return openapi_


@click.group()
def openapi():
    """OpenAPI management"""
    pass


@click.command()
@click.pass_context
@click.argument('config_file', type=click.File(encoding='utf-8'))
@click.option('--fail-on-invalid-collection/--no-fail-on-invalid-collection',
              '-fic', default=True, help='Fail on invalid collection')
@click.option('--format', '-f', 'format_', type=click.Choice(['json', 'yaml']),
              default='yaml', help='output format (json|yaml)')
@click.option('--output-file', '-of', type=click.File('w', encoding='utf-8'),
              help='Name of output file')
def generate(ctx, config_file, output_file, format_='yaml',
             fail_on_invalid_collection=True):
    """Generate OpenAPI Document"""

    if config_file is None:
        raise click.ClickException('--config/-c required')

    content = generate_openapi_document(
        config_file, format_, fail_on_invalid_collection)

    if output_file is None:
        click.echo(content)
    else:
        click.echo(f'Generating {output_file.name}')
        output_file.write(content)
        click.echo('Done')


@click.command()
@click.pass_context
@click.argument('openapi_file', type=click.File())
def validate(ctx, openapi_file):
    """Validate OpenAPI Document"""

    if openapi_file is None:
        raise click.ClickException('--openapi/-o required')

    click.echo(f'Validating {openapi_file}')
    instance = yaml_load(openapi_file)
    validate_openapi_document(instance)
    click.echo('Valid OpenAPI document')


openapi.add_command(generate)
openapi.add_command(validate)



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Angelos Tzotsos <tzotsos@gmail.com>
#
# Copyright (c) 2026 Tom Kralidis
# Copyright (c) 2025 Angelos Tzotsos
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging

import requests

from pygeoapi.pubsub.base import BasePubSubClient, PubSubClientConnectionError

LOGGER = logging.getLogger(__name__)


class HTTPPubSubClient(BasePubSubClient):
    """HTTP client"""

    def __init__(self, publisher_def):
        """
        Initialize object

        :param publisher_def: provider definition

        :returns: pygeoapi.pubsub.http.HTTPPubSubClient
        """

        super().__init__(publisher_def)
        self.name = 'HTTP'
        self.type = 'http'
        self.auth = None

        msg = f'Initializing to broker {self.broker_safe_url} with id {self.client_id}'  # noqa
        LOGGER.debug(msg)

        if None not in [self.broker_url.username, self.broker_url.password]:
            LOGGER.debug('Setting credentials')
            self.auth = (
                self.broker_url.username,
                self.broker_url.password
            )

    def connect(self) -> None:
        """
        Connect to an HTTP broker

        :returns: None
        """

        LOGGER.debug('No connection to HTTP')
        pass

    def pub(self, channel: str, message: str, qos: int = 1) -> bool:
        """
        Publish a message to a broker/channel

        :param channel: `str` of topic
        :param message: `str` of message

        :returns: `bool` of publish result
        """

        LOGGER.debug(f'Publishing to broker {self.broker_safe_url}')
        LOGGER.debug(f'Channel: {channel}')
        LOGGER.debug(f'Message: {message}')
        LOGGER.debug('Sanitizing channel for HTTP')
        channel = channel.replace('/', '-')
        channel = channel.replace(':', '-')
        LOGGER.debug(f'Sanitized channel for HTTP: {channel}')

        url = f'{self.broker}/{channel}'

        try:
            response = requests.post(url, auth=self.auth, json=message)
            response.raise_for_status()
        except Exception as err:
            raise PubSubClientConnectionError(err)

    def __repr__(self):
        return f'<HTTPPubSubClient> {self.broker_safe_url}'



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2026 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging

from pygeoapi.provider.base import ProviderTypeError

LOGGER = logging.getLogger(__name__)


def filter_providers_by_type(providers: list, type: str) -> dict:
    """
    helper function to filter a list of providers by type

    :param providers: ``list``
    :param type: str

    :returns: filtered ``dict`` provider
    """

    providers_ = {provider['type']: provider for provider in providers}
    return providers_.get(type)


def get_provider_by_type(providers: list, provider_type: str) -> dict:
    """
    helper function to load a provider by a provider type

    :param providers: ``list`` of providers
    :param provider_type: type of provider (e.g. feature)

    :returns: provider based on type
    """

    LOGGER.debug(f'Searching for provider type {provider_type}')
    try:
        p = (next(d for i, d in enumerate(providers)
                  if d['type'] == provider_type))
    except (RuntimeError, StopIteration):
        raise ProviderTypeError('Invalid provider type requested')

    return p


def get_provider_default(providers: list) -> dict:
    """
    helper function to get a resource's default provider

    :param providers: ``list`` of providers

    :returns: filtered ``dict``
    """

    try:
        default = (next(d for i, d in enumerate(providers) if 'default' in d
                   and d['default']))
        LOGGER.debug('found default provider type')
    except StopIteration:
        LOGGER.debug('no default provider type.  Returning first provider')
        default = providers[0]

    LOGGER.debug(f"Default provider: {default['type']}")
    return default



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Emmanuel Jolaiya <jolaiyaemmanuel@gmail.com>
#
# Copyright (c) 2024 Tom Kralidis
# Copyright (c) 2024 Emmanuel Jolaiya
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import enum
import logging

from typing import Any, Dict, List, Tuple, Union

from shapely.geometry import mapping, shape
from shapely.geometry.base import BaseGeometry
from shapely.prepared import prep
from shapely.wkt import loads

from pygeoapi.process.base import BaseProcessor, ProcessorExecuteError


LOGGER = logging.getLogger(__name__)


class SupportedShapelyOperations(enum.Enum):
    """
    Enum for the supported shapely operations
    """

    # Measurement operations.
    # Ref: https://shapely.readthedocs.io/en/stable/measurement.html

    MEASUREMENT_AREA = "measurement:area"
    MEASUREMENT_DISTANCE = "measurement:distance"
    MEASUREMENT_BOUNDS = "measurement:bounds"

    # Predicate operations.
    # Ref: https://shapely.readthedocs.io/en/stable/predicates.html

    PREDICATES_COVERS = "predicates:covers"
    PREDICATES_WITHIN = "predicates:within"

    # Set operations.
    # Ref: https://shapely.readthedocs.io/en/stable/set_operations.html

    SET_DIFFERENCE = "set:difference"
    SET_UNION = "set:union"

    # Constructive operations.
    # Ref: https://shapely.readthedocs.io/en/stable/constructive.html

    CONSTRUCTIVE_BUFFER = "constructive:buffer"
    CONSTRUCTIVE_CENTROID = "constructive:centroid"


class SupportedFormats(enum.Enum):
    WKT = "wkt"
    """Well Known Text"""
    GeoJSON_GEOMETRY = "geojson"
    """GeoJSON Geometry"""


class SupportedGeometryTypes(enum.Enum):
    POINT = "Point"
    POLYGON = "Polygon"
    MULTI_POLYGON = "MultiPolygon"


#: Process metadata and description
PROCESS_METADATA = {
    "version": "0.2.0",
    "id": "shapely-functions",
    "title": {
        "en": "Shapely Functions",
        "fr": "Fonctions galbées",
        "es": "Funciones bien formadas"
    },
    "description": {
        "en": "An example process that takes one or more input geometry "
        "(WKT or GeoJSON geometry), applies the specified shapely"
        " operation on the input, and returns the result "
        "in the specified output_format.",
        "fr": "Un exemple de processus qui prend une ou plusieurs géométries"
        " d'entrée (géométrie WKT ou GeoJSON) applique l'opération "
        "galbée spécifiée sur l'entrée et renvoie le "
        "résultat dans le format de sortie spécifié.",
        "es": "Un proceso de ejemplo que toma una o más geometrías"
        "de entrada (geometría WKT o GeoJSON) aplica la operación"
        "de forma especificada en la entrada y devuelve el resultado"
        "en el formato de salida especificado."
    },
    "jobControlOptions": ["sync-execute", "async-execute"],
    "keywords": [
        "shapely functions",
        "measurement",
        "predicates",
        "set operations",
        "constructive ops"
    ],
    "links": [
        {
            "type": "text/html",
            "rel": "about",
            "title": "information",
            "href": "https://shapely.readthedocs.io",
            "hreflang": "en-US"
        },
    ],
    "inputs": {
        "operation": {
            "title": "Shapely operation",
            "description": "The shapely operation to perform. Namespace of the"
            "function category is used to avoid mixup in function names.",
            "schema": {
                "type": "string"
            },
            "minOccurs": 1,
            "maxOccurs": 1,
            "enum": [ops.value for ops in SupportedShapelyOperations]
        },
        "geoms": {
            "title": "Input geometry",
            "description": "The representation of the object"
            " as GeoJSON geometry or a WKT.",
            "schema": {
                "type": "array",
                "description": "An array of the geometries. "
                "The order of the geometries matter. The first "
                "element is considered geom A, second element as geom B, "
                "in that order.",
                "minItems": 1,
                "maxItems": 2,
                "items": {
                    "oneOf": [
                        # Can be a WKT string
                        {
                            "type": "string",
                        },
                        # or a GeoJSON geometry
                        {
                            "oneOf": [
                                {"format": "geojson-geometry"},
                                {
                                    "$ref": "http://schemas.opengis.net/"
                                    "ogcapi/features/part1/1.0/openapi/"
                                    "schemas/geometryGeoJSON.yaml"
                                },
                            ]
                        },
                    ]
                },
            },
            "minOccurs": 1,
            "maxOccurs": 1,
        },
        "output_format": {
            "title": "The output format of the geometry",
            "description": "The output format of the process result. "
            "If the shapely operation does not return a geometry, then "
            " this is ignored.",
            "schema": {
                "type": "string"
            },
            "minOccurs": 0,
            "maxOccurs": 1,
            "enum": [v.value for v in SupportedFormats]
        },
    },
    "outputs": {
        "result": {
            "title": "Shapely operation result",
            "description": "The result of the shapely operation "
            "performed in the process.",
            "schema": {
                "type": "object",
                "contentMediaType": "application/json"
            },
        },
    },
    "example": {
        "inputs": {
            "operation": "predicates:within",
            "geoms": [
                {
                    "coordinates": [
                        [
                            [80.58983993887352, 24.07996699713047],
                            [80.58983993887352, 20.291958215654446],
                            [85.04987314197012, 20.291958215654446],
                            [85.04987314197012, 24.07996699713047],
                            [80.58983993887352, 24.07996699713047],
                        ]
                    ],
                    "type": "Polygon",
                },
                {
                    "coordinates": [
                        [
                            [80.58983993887352, 24.07996699713047],
                            [80.58983993887352, 20.291958215654446],
                            [85.04987314197012, 20.291958215654446],
                            [85.04987314197012, 24.07996699713047],
                            [80.58983993887352, 24.07996699713047],
                        ]
                    ],
                    "type": "Polygon"
                },
            ],
        }
    },
}


class ShapelyFunctionsProcessor(BaseProcessor):
    """Shapely Functions Processor example"""

    def __init__(self, processor_def):
        """
        Initialize object

        :param processor_def: provider definition

        :returns: pygeoapi.process.shapely_functions.ShapelyFunctionsProcessor
        """
        self.shapely_operations = [ops.value for ops in SupportedShapelyOperations]  # noqa: E501
        self.supported_formats = [fmt.value for fmt in SupportedFormats]
        super().__init__(processor_def, PROCESS_METADATA)

    def execute(self, data, outputs=None) -> Tuple[str, Dict[str, Any]]:
        mimetype = "application/json"
        operation = data.get("operation")
        output_format = data.get("output_format")
        geometries = data.get("geoms")

        # validations
        if geometries is None:
            raise ProcessorExecuteError("Cannot process without valid geometries.")  # noqa: E501

        if not isinstance(geometries, list):
            raise ProcessorExecuteError(
                """Cannot process without valid geometries.
                `geoms` must be an array of WKT
                 or GeoJSON geometry objects."""
            )

        # output_format could be optional,
        # because some operations return only value and not geometry
        nongeom_operations = {
            SupportedShapelyOperations.MEASUREMENT_AREA.value,
            SupportedShapelyOperations.MEASUREMENT_DISTANCE.value,
            SupportedShapelyOperations.MEASUREMENT_BOUNDS.value,
            SupportedShapelyOperations.PREDICATES_COVERS.value,
            SupportedShapelyOperations.PREDICATES_WITHIN.value
        }
        if output_format is None and operation not in nongeom_operations:
            raise ProcessorExecuteError("Cannot process without an `output_format`.")  # noqa: E501

        if (
            output_format is not None
            and output_format.lower() not in self.supported_formats
        ):
            raise ProcessorExecuteError(
                f"""Invalid `output_format` provided.
                Supported options are {self.supported_formats},
                but got {output_format}."""
            )

        self.output_format = output_format

        if operation is None:
            raise ProcessorExecuteError("Cannot process without a valid operation.")  # noqa: E501

        if operation not in self.shapely_operations:
            raise ProcessorExecuteError(
                f"""Invalid shapely operation provided.
                Supported operations are : {self.shapely_operations}"""
            )

        requires_single_geom = {
            SupportedShapelyOperations.MEASUREMENT_AREA.value,
            SupportedShapelyOperations.MEASUREMENT_BOUNDS.value,
            SupportedShapelyOperations.CONSTRUCTIVE_BUFFER.value,
            SupportedShapelyOperations.CONSTRUCTIVE_CENTROID.value
        }

        if operation in requires_single_geom and len(geometries) > 1:
            raise ProcessorExecuteError(
                f"""Too many geometries. The {operation}
                    operation requires only one geometry,
                    but {len(geometries)} was provided."""
            )

        if operation not in requires_single_geom and len(geometries) < 2:
            raise ProcessorExecuteError(
                f"""Too few geometries. The {operation} operation
                    requires at least two geometry,
                    but {len(geometries)} was provided."""
            )

        parsed_geoms = []

        for geom in geometries:
            if isinstance(geom, str):
                parsed_geoms.append(loads(geom))
            else:
                parsed_geoms.append(shape(geom))

        result = self.perform_operation(parsed_geoms, operation)

        return mimetype, result

    def parse_result(self, geom: BaseGeometry) -> Union[str, Dict[str, Any]]:
        """
        Convert a shapely geometry into the specified format by the client.

        Args:
            geom (BaseGeometry): The shapely geometry to convert.

        Returns:
            Union[str,Dict[str,Any]]: The resulting geometry
            in the specified format.
        """
        return (
            geom.wkt
            if self.output_format == SupportedFormats.WKT.value
            else mapping(geom)
        )

    def perform_operation(self, parsed_geoms: List[BaseGeometry], operation: str):  # noqa: E501
        """
        Perform the exact shapely operation specified by the client.

        Args:
            parsed_geoms (List[BaseGeometry]): An array of shapely geometries.
            operation (str): The exact shapely operation to perform.

        Raises:
            ProcessorExecuteError: Server error if given an invalid parameter.

        Returns:
            Dict[str,Any]: The response object with the operation result,
            to be returned to the client.
        """
        result = {"operation": operation, "result": None}

        #  prep the first geom for performance improvement
        prep(parsed_geoms[0])
        if operation == SupportedShapelyOperations.MEASUREMENT_AREA.value:
            if parsed_geoms[0].type not in [
                SupportedGeometryTypes.POLYGON.value,
                SupportedGeometryTypes.MULTI_POLYGON.value
            ]:
                raise ProcessorExecuteError(
                    f"""Invalid geometry type.{operation}
                        operation only works on
                        `Polygon and MultiPolygon` geometry."""
                )
            result.update({"result": parsed_geoms[0].area})
        elif operation == SupportedShapelyOperations.MEASUREMENT_BOUNDS.value:
            result.update({"result": parsed_geoms[0].bounds})
        elif operation == SupportedShapelyOperations.MEASUREMENT_DISTANCE.value:  # noqa: E501
            result.update({"result": parsed_geoms[0].distance(parsed_geoms[1])})  # noqa: E501
        elif operation == SupportedShapelyOperations.PREDICATES_COVERS.value:
            result.update({"result": parsed_geoms[0].covers(parsed_geoms[1])})
        elif operation == SupportedShapelyOperations.PREDICATES_WITHIN.value:
            result.update({"result": parsed_geoms[0].within(parsed_geoms[1])})
        elif operation == SupportedShapelyOperations.SET_DIFFERENCE.value:
            result.update(
                {
                    "result": self.parse_result(
                        parsed_geoms[0].difference(parsed_geoms[1])
                    )
                }
            )
        elif operation == SupportedShapelyOperations.SET_UNION.value:
            result.update(
                {
                    "result": self.parse_result(parsed_geoms[0].union(parsed_geoms[1]))  # noqa: E501
                }
            )
        elif operation == SupportedShapelyOperations.CONSTRUCTIVE_BUFFER.value:
            # todo - how do we receive kwargs from the user?
            result.update({"result": self.parse_result(parsed_geoms[0].buffer(10))})  # noqa: E501
        else:
            result.update({"result": self.parse_result(parsed_geoms[0].centroid)})  # noqa: E501
        return result

    def __repr__(self):
        return f"<ShapelyFunctionsProcessor> {self.name}"



# =================================================================

# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2026 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from datetime import datetime, UTC
import json
import logging
import uuid
from typing import Union

LOGGER = logging.getLogger(__name__)

CONFORMANCE_CLASSES = [
    'https://www.opengis.net/spec/ogcapi-pubsub-1/1.0/conf/message-payload-cloudevents-json',  # noqa
    'https://www.opengis.net/spec/ogcapi-pubsub-1/1.0/conf/discovery'
]


def publish_message(pubsub_client, url: str, action: str,
                    resource: str = None, item: str = None,
                    data: dict = None) -> bool:
    """
    Publish broker message

    :param pubsub_client: `pygeoapi.pubsub.BasePubSubClient` instance
    :param url: `str` of server base URL
    :param action: `str` of action trigger name (create, update, delete)
    :param resource: `str` of resource identifier
    :param item: `str` of item identifier
    :param data: `dict` of data payload

    :returns: `bool` of whether message publishing was successful
    """

    if action in ['create', 'update']:
        channel = f'collections/{resource}'
        data_ = data
        media_type = 'application/geo+json'
        type_ = f'org.ogc.api.collection.item.{action}'
    elif action == 'delete':
        channel = f'collections/{resource}'
        data_ = item
        media_type = 'text/plain'
        type_ = f'org.ogc.api.collection.item.{action}'
    elif action == 'process':
        channel = f'processes/{resource}'
        media_type = 'application/json'
        data_ = data
        type_ = 'org.ogc.api.job.result'

    if pubsub_client.channel is not None:
        channel = f'{pubsub_client.channel}/{channel}'

    message = generate_ogc_cloudevent(type_, media_type, url,
                                      channel, data_)
    LOGGER.debug(f'Message: {message}')

    try:
        pubsub_client.connect()
        pubsub_client.pub(channel, json.dumps(message))
    except Exception as err:
        raise RuntimeError(err)


def generate_ogc_cloudevent(type_: str, media_type: str, source: str,
                            subject: str, data: Union[dict, str]) -> dict:
    """
    Generate CloudEvent

    :param type_: `str` of CloudEvents type
    :param source: `str` of source
    :param subject: `str` of subject
    :param media_type: `str` of media type
    :param data: `str` or `dict` of data

    :returns: `dict` of OGC CloudEvent payload
    """

    try:
        data2 = json.loads(data)
    except Exception:
        if isinstance(data, bytes):
            data2 = data.decode('utf-8')
        else:
            data2 = data

    message = {
        'specversion': '1.0',
        'type': type_,
        'source': source,
        'subject': subject,
        'id': str(uuid.uuid4()),
        'time': datetime.now(UTC).strftime('%Y-%m-%dT%H:%M:%SZ'),
        'datacontenttype': media_type,
        # 'dataschema': 'TODO',
        'data': data2
    }

    return message


def get_oas_30(cfg, locale_):
    return [], {}



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Francesco Bartoli <xbartolone@gmail.com>
#          Sander Schaminee <sander.schaminee@geocat.net>
#          John A Stevenson <jostev@bgs.ac.uk>
#          Colin Blackburn <colb@bgs.ac.uk>
#          Ricardo Garcia Silva <ricardo.garcia.silva@geobeyond.it>
#
# Copyright (c) 2026 Tom Kralidis
# Copyright (c) 2026 Francesco Bartoli
# Copyright (c) 2022 John A Stevenson and Colin Blackburn
# Copyright (c) 2023 Ricardo Garcia Silva
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from copy import deepcopy
import logging

from pygeoapi import l10n
from pygeoapi.formats import (F_JSON, F_JSONLD, F_HTML, F_JPEG,
                              F_PNG, FORMAT_TYPES)
from pygeoapi.crs import DEFAULT_STORAGE_CRS, get_supported_crs_list
from pygeoapi.plugin import load_plugin
from pygeoapi.provider import get_provider_by_type, get_provider_default
from pygeoapi.provider.base import ProviderConnectionError, ProviderTypeError
from pygeoapi.util import dategetter, get_dataset_formatters

LOGGER = logging.getLogger(__name__)

OGC_RELTYPES_BASE = 'http://www.opengis.net/def/rel/ogc/1.0'


def gen_collection(api, request, dataset: str,
                   locale_: str) -> dict:
    """
    Generate OGC API Collection description

    :param api: `APIRequest` object
    :param dataset: `str` of dataset name
    :param locale_: `str` of requested locale

    :returns: `dict` of OGC API Collection description
    """

    config = api.config['resources'][dataset]

    data = {
        'id': dataset,
        'links': []
    }

    collection_data = get_provider_default(config['providers'])
    collection_data_type = collection_data['type']

    collection_data_format = None

    if 'format' in collection_data:
        collection_data_format = collection_data['format']

    is_vector_tile = (collection_data_type == 'tile' and
                      collection_data_format['name'] not
                      in [F_PNG, F_JPEG])

    data.update({
        'title': l10n.translate(config['title'], locale_),
        'description': l10n.translate(config['description'], locale_),
        'keywords': l10n.translate(config['keywords'], locale_),
    })

    extents = deepcopy(config['extents'])

    bbox = extents['spatial']['bbox']
    LOGGER.debug('Setting spatial extents from configuration')
    # The output should be an array of bbox, so if the user only
    # provided a single bbox, wrap it in a array.
    if not isinstance(bbox[0], list):
        bbox = [bbox]

    data['extent'] = {
        'spatial': {
            'bbox': bbox
        }
    }

    if 'crs' in extents['spatial']:
        data['extent']['spatial']['crs'] = extents['spatial']['crs']

    t_ext = extents.get('temporal', {})
    if t_ext:
        LOGGER.debug('Setting temporal extents from configuration')
        begins = dategetter('begin', t_ext)
        ends = dategetter('end', t_ext)
        data['extent']['temporal'] = {
            'interval': [[begins, ends]]
        }
        if 'trs' in t_ext:
            data['extent']['temporal']['trs'] = t_ext['trs']
        if 'resolution' in t_ext:
            data['extent']['temporal']['grid'] = {
                'resolution': t_ext['resolution']
            }
        if 'default' in t_ext:
            data['extent']['temporal']['default'] = t_ext['default']

    _ = extents.pop('spatial', None)
    _ = extents.pop('temporal', None)

    for ek, ev in extents.items():
        LOGGER.debug(f'Adding extent {ek}')
        data['extent'][ek] = {
            'definition': ev['url'],
            'interval': [ev['range']]
        }
        if 'units' in ev:
            data['extent'][ek]['unit'] = ev['units']

        if 'values' in ev:
            data['extent'][ek]['grid'] = {
                'cellsCount': len(ev['values']),
                'coordinates': ev['values']
            }

    LOGGER.debug('Processing configured collection links')
    for link in l10n.translate(config.get('links', []), locale_):
        lnk = {
            'type': link['type'],
            'rel': link['rel'],
            'title': l10n.translate(link['title'], locale_),
            'href': l10n.translate(link['href'], locale_),
        }
        if 'hreflang' in link:
            lnk['hreflang'] = l10n.translate(
                link['hreflang'], locale_)
        content_length = link.get('length', 0)

        if lnk['rel'] == 'enclosure' and content_length == 0:
            # Issue HEAD request for enclosure links without length
            lnk_headers = api.prefetcher.get_headers(lnk['href'])
            content_length = int(lnk_headers.get('content-length', 0))
            content_type = lnk_headers.get('content-type', lnk['type'])
            if content_length == 0:
                # Skip this (broken) link
                LOGGER.debug(f"Enclosure {lnk['href']} is invalid")
                continue
            if content_type != lnk['type']:
                # Update content type if different from specified
                lnk['type'] = content_type
                LOGGER.debug(
                    f"Fixed media type for enclosure {lnk['href']}")

        if content_length > 0:
            lnk['length'] = content_length

        data['links'].append(lnk)

    # TODO: provide translations
    LOGGER.debug('Adding JSON and HTML link relations')
    data['links'].extend([{
        'type': FORMAT_TYPES[F_JSON],
        'rel': 'root',
        'title': l10n.translate('The landing page of this server as JSON', locale_),  # noqa
        'href': f"{api.base_url}?f={F_JSON}"
    }, {
        'type': FORMAT_TYPES[F_HTML],
        'rel': 'root',
        'title': l10n.translate('The landing page of this server as HTML', locale_),  # noqa
        'href': f"{api.base_url}?f={F_HTML}"
    }, {
        'type': FORMAT_TYPES[F_JSON],
        'rel': request.get_linkrel(F_JSON),
        'title': l10n.translate('This document as JSON', locale_),
        'href': f'{api.get_collections_url()}/{dataset}?f={F_JSON}'
    }, {
        'type': FORMAT_TYPES[F_JSONLD],
        'rel': request.get_linkrel(F_JSONLD),
        'title': l10n.translate('This document as RDF (JSON-LD)', locale_),
        'href': f'{api.get_collections_url()}/{dataset}?f={F_JSONLD}'
    }, {
        'type': FORMAT_TYPES[F_HTML],
        'rel': request.get_linkrel(F_HTML),
        'title': l10n.translate('This document as HTML', locale_),
        'href': f'{api.get_collections_url()}/{dataset}?f={F_HTML}'
    }])

    if collection_data_type == 'record':
        data['links'].extend([{
            'type': FORMAT_TYPES[F_JSON],
            'rel': f'{OGC_RELTYPES_BASE}/ogc-catalog',
            'title': l10n.translate('Record catalogue as JSON', locale_),
            'href': f'{api.get_collections_url()}/{dataset}?f={F_JSON}'
        }, {
            'type': FORMAT_TYPES[F_HTML],
            'rel': f'{OGC_RELTYPES_BASE}/ogc-catalog',
            'title': l10n.translate('Record catalogue as HTML', locale_),
            'href': f'{api.get_collections_url()}/{dataset}?f={F_HTML}'
        }])

    if collection_data_type in ['feature', 'coverage', 'record']:
        data['links'].extend([{
            'type': 'application/schema+json',
            'rel': f'{OGC_RELTYPES_BASE}/schema',
            'title': l10n.translate('Schema of collection in JSON', locale_),
            'href': f'{api.get_collections_url()}/{dataset}/schema?f={F_JSON}'
        }, {
            'type': FORMAT_TYPES[F_HTML],
            'rel': f'{OGC_RELTYPES_BASE}/schema',
            'title': l10n.translate('Schema of collection in HTML', locale_),
            'href': f'{api.get_collections_url()}/{dataset}/schema?f={F_HTML}'
        }])

    if is_vector_tile or collection_data_type in ['feature', 'record']:
        # TODO: translate
        data['itemType'] = collection_data_type
        LOGGER.debug('Adding feature/record based links')
        data['links'].extend([{
            'type': 'application/schema+json',
            'rel': f'{OGC_RELTYPES_BASE}/queryables',
            'title': l10n.translate('Queryables for this collection as JSON', locale_),  # noqa
            'href': f'{api.get_collections_url()}/{dataset}/queryables?f={F_JSON}'  # noqa
        }, {
            'type': FORMAT_TYPES[F_HTML],
            'rel': f'{OGC_RELTYPES_BASE}/queryables',
            'title': l10n.translate('Queryables for this collection as HTML', locale_),  # noqa
            'href': f'{api.get_collections_url()}/{dataset}/queryables?f={F_HTML}'  # noqa
        }, {
            'type': 'application/geo+json',
            'rel': 'items',
            'title': l10n.translate('Items as GeoJSON', locale_),
            'href': f'{api.get_collections_url()}/{dataset}/items?f={F_JSON}'
        }, {
            'type': FORMAT_TYPES[F_JSONLD],
            'rel': 'items',
            'title': l10n.translate('Items as RDF (GeoJSON-LD)', locale_),
            'href': f'{api.get_collections_url()}/{dataset}/items?f={F_JSONLD}'
        }, {
            'type': FORMAT_TYPES[F_HTML],
            'rel': 'items',
            'title': l10n.translate('Items as HTML', locale_),  # noqa
            'href': f'{api.get_collections_url()}/{dataset}/items?f={F_HTML}'
        }])

        for key, value in get_dataset_formatters(config).items():
            data['links'].append({
                'type': value.mimetype,
                'rel': 'items',
                'title': l10n.translate(f'Items as {key}', locale_),  # noqa
                'href': f'{api.get_collections_url()}/{dataset}/items?f={value.f}'  # noqa
            })

    # OAPIF Part 2 - list supported CRSs and StorageCRS
    if collection_data_type in ['edr', 'feature']:
        data['crs'] = get_supported_crs_list(collection_data)
        data['storageCrs'] = collection_data.get('storage_crs', DEFAULT_STORAGE_CRS)  # noqa
        if 'storage_crs_coordinate_epoch' in collection_data:
            data['storageCrsCoordinateEpoch'] = collection_data.get('storage_crs_coordinate_epoch')  # noqa

    elif collection_data_type == 'coverage':
        LOGGER.debug('Adding coverage based links')
        data['links'].append({
            'type': 'application/prs.coverage+json',
            'rel': f'{OGC_RELTYPES_BASE}/coverage',
            'title': l10n.translate('Coverage data', locale_),
            'href': f'{api.get_collections_url()}/{dataset}/coverage?f={F_JSON}'  # noqa
        })
        if collection_data_format is not None:
            title_ = l10n.translate('Coverage data as', locale_)
            title_ = f"{title_} {collection_data_format['name']}"
            data['links'].append({
                'type': collection_data_format['mimetype'],
                'rel': f'{OGC_RELTYPES_BASE}/coverage',
                'title': title_,
                'href': f"{api.get_collections_url()}/{dataset}/coverage?f={collection_data_format['name']}"  # noqa
            })
        if dataset is not None:
            LOGGER.debug('Creating extended coverage metadata')
            try:
                provider_def = get_provider_by_type(
                    api.config['resources'][dataset]['providers'],
                    'coverage')
                p = load_plugin('provider', provider_def)
            except ProviderConnectionError:
                raise
            except ProviderTypeError:
                pass
            else:
                data['extent']['spatial']['grid'] = [{
                    'cellsCount': p._coverage_properties['width'],
                    'resolution': p._coverage_properties['resx']
                }, {
                    'cellsCount': p._coverage_properties['height'],
                    'resolution': p._coverage_properties['resy']
                }]
                if 'time_range' in p._coverage_properties:
                    data['extent']['temporal'] = {
                        'interval': [p._coverage_properties['time_range']]
                    }
                    if 'restime' in p._coverage_properties:
                        data['extent']['temporal']['grid'] = {
                            'resolution': p._coverage_properties['restime']
                        }
                if 'uad' in p._coverage_properties:
                    data['extent'].update(p._coverage_properties['uad'])

    try:
        tile = get_provider_by_type(config['providers'], 'tile')
        p = load_plugin('provider', tile)
    except ProviderConnectionError:
        raise
    except ProviderTypeError:
        tile = None

    if tile:
        LOGGER.debug('Adding tile links')
        data['links'].extend([{
            'type': FORMAT_TYPES[F_JSON],
            'rel': f'{OGC_RELTYPES_BASE}/tilesets-{p.tile_type}',
            'title': l10n.translate('Tiles as JSON', locale_),
            'href': f'{api.get_collections_url()}/{dataset}/tiles?f={F_JSON}'
        }, {
            'type': FORMAT_TYPES[F_HTML],
            'rel': f'{OGC_RELTYPES_BASE}/tilesets-{p.tile_type}',
            'title': l10n.translate('Tiles as HTML', locale_),
            'href': f'{api.get_collections_url()}/{dataset}/tiles?f={F_HTML}'
        }])

    try:
        map_ = get_provider_by_type(config['providers'], 'map')
        p = load_plugin('provider', map_)
    except ProviderTypeError:
        map_ = None

    if map_:
        LOGGER.debug('Adding map links')

        map_mimetype = map_['format']['mimetype']
        map_format = map_['format']['name']

        title_ = l10n.translate('Map as', locale_)
        title_ = f'{title_} {map_format}'

        data['links'].append({
            'type': map_mimetype,
            'rel': f'{OGC_RELTYPES_BASE}/map',
            'title': title_,
            'href': f'{api.get_collections_url()}/{dataset}/map?f={map_format}'
        })

        if p._fields:
            schema_reltype = f'{OGC_RELTYPES_BASE}/schema',
            schema_links = [s for s in data['links'] if
                            schema_reltype in s]

            if not schema_links:
                title_ = l10n.translate('Schema of collection in JSON', locale_)  # noqa
                data['links'].append({
                    'type': 'application/schema+json',
                    'rel': f'{OGC_RELTYPES_BASE}/schema',
                    'title': title_,
                    'href': f'{api.get_collections_url()}/{dataset}/schema?f=json'  # noqa
                })
                title_ = l10n.translate('Schema of collection in HTML', locale_)  # noqa
                data['links'].append({
                    'type': 'text/html',
                    'rel': f'{OGC_RELTYPES_BASE}/schema',
                    'title': title_,
                    'href': f'{api.get_collections_url()}/{dataset}/schema?f=html'  # noqa
                })

    try:
        edr = get_provider_by_type(config['providers'], 'edr')
        p = load_plugin('provider', edr)
    except ProviderConnectionError:
        raise
    except ProviderTypeError:
        edr = None

    if edr:
        # TODO: translate
        LOGGER.debug('Adding EDR links')
        data['data_queries'] = {}
        parameters = p.get_fields()
        if parameters:
            data['parameter_names'] = {}
            for key, value in parameters.items():
                data['parameter_names'][key] = {
                    'id': key,
                    'type': 'Parameter',
                    'name': value['title'],
                    'observedProperty': {
                        'label': {
                            'id': key,
                            'en': value['title']
                        },
                    },
                    'unit': {
                        'label': {
                            'en': value['title']
                        },
                        'symbol': {
                            'value': value['x-ogc-unit'],
                            'type': 'http://www.opengis.net/def/uom/UCUM/'
                        }
                    }
                }

                data['parameter_names'][key].update({
                    'description': value['description']}
                        if 'description' in value else {}
                )

        for qt in p.get_query_types():
            data_query = {
                'link': {
                    'href': f'{api.get_collections_url()}/{dataset}/{qt}',
                    'rel': 'data',
                    'variables': {
                        'query_type': qt
                    }
                }
            }

            if request.format is not None and request.format == 'json':
                data_query['link']['type'] = 'application/vnd.cov+json'

            data['data_queries'][qt] = data_query

            title1 = l10n.translate('query for this collection as JSON', locale_)  # noqa
            title1 = f'{qt} {title1}'
            title2 = l10n.translate('query for this collection as HTML', locale_)  # noqa
            title2 = f'{qt} {title2}'

            data['links'].extend([{
                'type': 'application/json',
                'rel': 'data',
                'title': title1,
                'href': f'{api.get_collections_url()}/{dataset}/{qt}?f={F_JSON}'  # noqa
            }, {
                'type': FORMAT_TYPES[F_HTML],
                'rel': 'data',
                'title': title2,
                'href': f'{api.get_collections_url()}/{dataset}/{qt}?f={F_HTML}'  # noqa 
            }])

            for key, value in get_dataset_formatters(config).items():
                title3 = f'{qt} query for this collection as {key}'
                data['links'].append({
                    'type': value.mimetype,
                    'rel': 'data',
                    'title': title3,
                    'href': f'{api.get_collections_url()}/{dataset}/{qt}?f={value.f}'  # noqa
                })

    return data



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2026 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging

from pygeoapi.provider.base import (BaseProvider, ProviderInvalidDataError,
                                    ProviderQueryError)

LOGGER = logging.getLogger(__name__)

EDR_QUERY_TYPES = ['position', 'radius', 'area', 'cube',
                   'trajectory', 'corridor', 'items',
                   'locations', 'instances']


class BaseEDRProvider(BaseProvider):
    """Base EDR Provider"""

    query_types = []

    def __init__(self, provider_def):
        """
        Initialize object

        :param provider_def: provider definition

        :returns: pygeoapi.provider.base_edr.BaseEDRProvider
        """

        BaseProvider.__init__(self, provider_def)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        cls.query_types = [
            name for name, function in cls.__dict__.items()
            if name in EDR_QUERY_TYPES and callable(function)
        ]

        if not cls.query_types:
            msg = f"{cls.__name__} does not implement any query types"
            LOGGER.error(msg)
            raise ProviderInvalidDataError(msg)

        LOGGER.debug(
            f'{cls.__name__} registered query types: {cls.query_types}'
        )

        if 'items' in cls.query_types:
            LOGGER.warning(
                f'items query is registered in {cls.__name__}, '
                'but requests will be routed to a feature provider'
            )

    def get_instances(self):
        """
        Get a list of instance identifiers

        :returns: `list` of instance identifiers
        """

        return NotImplementedError()

    def get_instance(self, instance):
        """
        Validate instance identifier

        :returns: `bool` of whether instance is valid
        """

        return NotImplementedError()

    def get_query_types(self):
        """
        Provide supported query types

        :returns: `list` of EDR query types
        """

        return self.query_types

    def query(self, **kwargs):
        """
        Extract data from collection collection

        :param query_type: query type
        :param wkt: `shapely.geometry` WKT geometry
        :param datetime_: temporal (datestamp or extent)
        :param select_properties: list of parameters
        :param z: vertical level(s)
        :param format_: data format of output
        :param bbox: bbox geometry (for cube queries)
        :param within: distance (for radius queries)
        :param within_units: distance units (for radius queries)
        :param instance: instance name (for instances queries)
        :param limit: number of records to return (for locations queries)
        :param location_id: location identifier (for locations queries)

        :returns: coverage data as `dict` of CoverageJSON or native format
        """
        query_type = kwargs.get('query_type')
        if query_type is None:
            raise ProviderQueryError('Query type is required')
        try:
            query_function = getattr(self, query_type)
        except AttributeError:
            raise ProviderQueryError('Query type not implemented')

        return query_function(**kwargs)



# =================================================================
#
# Authors: Francesco Bartoli <xbartolone@gmail.com>
#          Tom Kralidis <tomkralidis@gmail.com>
#          Abdulazeez Abdulazeez Adeshina <youngestdev@gmail.com>
#
# Copyright (c) 2025 Francesco Bartoli
# Copyright (c) 2026 Tom Kralidis
# Copyright (c) 2022 Abdulazeez Abdulazeez Adeshina
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================
""" Starlette module providing the route paths to the api"""

import asyncio
from http import HTTPStatus
import os
from typing import Callable, Union
from pathlib import Path

import click
from starlette.routing import Route, Mount
from starlette.staticfiles import StaticFiles
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.datastructures import URL
from starlette.types import ASGIApp, Scope, Send, Receive
from starlette.responses import (
    Response, JSONResponse, HTMLResponse, RedirectResponse
)
import uvicorn

from pygeoapi.api import API, APIRequest, apply_gzip
import pygeoapi.api as core_api
import pygeoapi.api.coverages as coverages_api
import pygeoapi.api.environmental_data_retrieval as edr_api
import pygeoapi.api.itemtypes as itemtypes_api
import pygeoapi.api.maps as maps_api
import pygeoapi.api.processes as processes_api
import pygeoapi.api.stac as stac_api
import pygeoapi.api.tiles as tiles_api
from pygeoapi.asyncapi import load_asyncapi_document
from pygeoapi.openapi import load_openapi_document
from pygeoapi.config import get_config
from pygeoapi.util import get_api_rules

CONFIG = get_config()

if 'PYGEOAPI_OPENAPI' not in os.environ:
    raise RuntimeError('PYGEOAPI_OPENAPI environment variable not set')

OPENAPI = load_openapi_document()
ASYNCAPI = load_asyncapi_document()

if CONFIG['server'].get('admin'):
    import pygeoapi.api.admin as admin_api
    from pygeoapi.api.admin import Admin

p = Path(__file__)

APP = Starlette(debug=True)
STATIC_DIR = Path(p).parent.resolve() / 'static'

try:
    STATIC_DIR = Path(CONFIG['server']['templates']['static'])
except KeyError:
    pass

API_RULES = get_api_rules(CONFIG)

api_ = API(CONFIG, OPENAPI, ASYNCAPI)


def call_api_threadsafe(
    loop: asyncio.AbstractEventLoop, api_call: Callable, *args
) -> tuple:
    """
    The api call needs a running loop. This method is meant to be called
    from a thread that has no loop running.

    :param loop: The loop to use.
    :param api_call: The API method to call.
    :param args: Arguments to pass to the API method.
    :returns: The api call result tuple.
    """
    asyncio.set_event_loop(loop)
    return api_call(*args)


def _to_response(headers: dict, status: int,
                 content: str | bytes) -> Response:
    if headers['Content-Type'] == 'text/html':
        response = HTMLResponse(content=content, status_code=status)
    else:
        if isinstance(content, dict):
            response = JSONResponse(content, status_code=status)
        else:
            response = Response(content, status_code=status)

    if headers is not None:
        response.headers.update(headers)
    return response


async def execute_from_starlette(api_function, request: Request, *args,
                                 skip_valid_check: bool = False,
                                 alternative_api: API | None = None
                                 ) -> Response:
    actual_api = api_ if alternative_api is None else alternative_api
    api_request = await APIRequest.from_starlette(request, actual_api.locales)
    content: Union[str, bytes]
    if not skip_valid_check and not api_request.is_valid():
        headers, status, content = actual_api.get_format_exception(api_request)
    else:

        loop = asyncio.get_running_loop()
        headers, status, content = await loop.run_in_executor(
            None, call_api_threadsafe, loop, api_function,
            actual_api, api_request, *args)
        # 204 responses must have an empty body, but gzip
        # encoding would add gzip metadata, thus we skip
        if status != HTTPStatus.NO_CONTENT:
            content = apply_gzip(headers, content)

    response = _to_response(headers, status, content)

    return response


async def landing_page(request: Request) -> Response:
    """
    OGC API landing page endpoint

    :param request: Starlette Request instance

    :returns: Starlette HTTP Response
    """
    return await execute_from_starlette(core_api.landing_page, request)


async def openapi(request: Request) -> Response:
    """
    OpenAPI endpoint

    :param request: Starlette Request instance

    :returns: Starlette HTTP Response
    """
    return await execute_from_starlette(core_api.openapi_, request)


async def asyncapi(request: Request) -> Response:
    """
    AsyncAPI endpoint

    :param request: Starlette Request instance

    :returns: Starlette HTTP Response
    """
    return await execute_from_starlette(core_api.asyncapi_, request)


async def conformance(request: Request) -> Response:
    """
    OGC API conformance endpoint

    :param request: Starlette Request instance

    :returns: Starlette HTTP Response
    """
    return await execute_from_starlette(core_api.conformance, request)


async def get_tilematrix_set(request: Request,
                             tileMatrixSetId: str | None = None) -> Response:
    """
    OGC API TileMatrixSet endpoint

    :param tileMatrixSetId: identifier of tile matrix set
    :returns: HTTP response
    """
    if 'tileMatrixSetId' in request.path_params:
        tileMatrixSetId = request.path_params['tileMatrixSetId']

    return await execute_from_starlette(
        tiles_api.tilematrixset, request, tileMatrixSetId,
    )


async def get_tilematrix_sets(request: Request) -> Response:
    """
    OGC API TileMatrixSets endpoint

    :returns: HTTP response
    """
    return await execute_from_starlette(tiles_api.tilematrixsets, request)


async def collection_schema(request: Request,
                            collection_id: str | None = None) -> Response:
    """
    OGC API collections schema endpoint

    :param request: Starlette Request instance
    :param collection_id: collection identifier

    :returns: Starlette HTTP Response
    """
    if 'collection_id' in request.path_params:
        collection_id = request.path_params['collection_id']

    return await execute_from_starlette(core_api.get_collection_schema,
                                        request, collection_id)


async def collection_queryables(request: Request,
                                collection_id: str | None = None) -> Response:
    """
    OGC API collections queryables endpoint

    :param request: Starlette Request instance
    :param collection_id: collection identifier

    :returns: Starlette HTTP Response
    """
    if 'collection_id' in request.path_params:
        collection_id = request.path_params['collection_id']

    return await execute_from_starlette(
        itemtypes_api.get_collection_queryables, request, collection_id,
    )


async def get_collection_tiles(request: Request,
                               collection_id: str | None = None) -> Response:
    """
    OGC open api collections tiles access point

    :param request: Starlette Request instance
    :param collection_id: collection identifier

    :returns: Starlette HTTP Response
    """
    if 'collection_id' in request.path_params:
        collection_id = request.path_params['collection_id']

    return await execute_from_starlette(
        tiles_api.get_collection_tiles, request, collection_id)


async def get_collection_tiles_metadata(request: Request,
                                        collection_id: str | None = None,
                                        tileMatrixSetId: str | None = None):
    """
    OGC open api collection tiles service metadata

    :param collection_id: collection identifier
    :param tileMatrixSetId: identifier of tile matrix set

    :returns: HTTP response
    """
    if 'collection_id' in request.path_params:
        collection_id = request.path_params['collection_id']
    if 'tileMatrixSetId' in request.path_params:
        tileMatrixSetId = request.path_params['tileMatrixSetId']

    return await execute_from_starlette(
        tiles_api.get_collection_tiles_metadata, request,
        collection_id, tileMatrixSetId, skip_valid_check=True,
    )


async def get_collection_items_tiles(request: Request,
                                     collection_id: str | None = None,
                                     tileMatrixSetId: str | None = None,
                                     tile_matrix: str | None = None,
                                     tileRow: str | None = None,
                                     tileCol: str | None = None) -> Response:
    """
    OGC open api collection tiles service

    :param request: Starlette Request instance
    :param collection_id: collection identifier
    :param tileMatrixSetId: identifier of tile matrix set
    :param tile_matrix: identifier of {z} matrix index
    :param tileRow: identifier of {y} matrix index
    :param tileCol: identifier of {x} matrix index

    :returns: HTTP response
    """
    if 'collection_id' in request.path_params:
        collection_id = request.path_params['collection_id']
    if 'tileMatrixSetId' in request.path_params:
        tileMatrixSetId = request.path_params['tileMatrixSetId']
    if 'tile_matrix' in request.path_params:
        tile_matrix = request.path_params['tile_matrix']
    if 'tileRow' in request.path_params:
        tileRow = request.path_params['tileRow']
    if 'tileCol' in request.path_params:
        tileCol = request.path_params['tileCol']
    return await execute_from_starlette(
        tiles_api.get_collection_tiles_data, request, collection_id,
        tileMatrixSetId, tile_matrix, tileRow, tileCol,
        skip_valid_check=True,
    )


async def collection_items(request: Request,
                           collection_id: str | None = None,
                           item_id: str | None = None) -> Response:
    """
    OGC API collections items endpoint

    :param request: Starlette Request instance
    :param collection_id: collection identifier
    :param item_id: item identifier

    :returns: Starlette HTTP Response
    """

    if 'collection_id' in request.path_params:
        collection_id = request.path_params['collection_id']
    if 'item_id' in request.path_params:
        item_id = request.path_params['item_id']
    if item_id is None:
        if request.method == 'POST':  # filter or manage items
            content_type = request.headers.get('content-type')
            if content_type is not None:
                if content_type == 'application/geo+json':
                    return await execute_from_starlette(
                        itemtypes_api.manage_collection_item, request,
                        'create', collection_id, skip_valid_check=True)
                else:
                    return await execute_from_starlette(
                        itemtypes_api.get_collection_items,
                        request,
                        collection_id,
                        skip_valid_check=True,
                    )
        elif request.method == 'OPTIONS':
            return await execute_from_starlette(
                itemtypes_api.manage_collection_item, request,
                'options', collection_id, skip_valid_check=True,
            )
        else:  # GET: list items
            return await execute_from_starlette(
                itemtypes_api.get_collection_items, request, collection_id,
                skip_valid_check=True)

    elif request.method == 'DELETE':
        return await execute_from_starlette(
            itemtypes_api.manage_collection_item, request, 'delete',
            collection_id, item_id, skip_valid_check=True,
        )
    elif request.method == 'PUT':
        return await execute_from_starlette(
            itemtypes_api.manage_collection_item, request, 'update',
            collection_id, item_id, skip_valid_check=True,
        )
    elif request.method == 'OPTIONS':
        return await execute_from_starlette(
            itemtypes_api.manage_collection_item, request, 'options',
            collection_id, item_id, skip_valid_check=True,
        )
    else:
        return await execute_from_starlette(
            itemtypes_api.get_collection_item, request, collection_id, item_id)


async def collection_coverage(request: Request,
                              collection_id: str | None = None) -> Response:
    """
    OGC API - Coverages coverage endpoint

    :param request: Starlette Request instance
    :param collection_id: collection identifier

    :returns: Starlette HTTP Response
    """
    if 'collection_id' in request.path_params:
        collection_id = request.path_params['collection_id']

    return await execute_from_starlette(
        coverages_api.get_collection_coverage, request, collection_id,
        skip_valid_check=True)


async def collection_map(request: Request,
                         collection_id: str | None = None,
                         style_id: str | None = None) -> Response:
    """
    OGC API - Maps map render endpoint

    :param collection_id: collection identifier
    :param style_id: style identifier

    :returns: HTTP response
    """

    if 'collection_id' in request.path_params:
        collection_id = request.path_params['collection_id']
    if 'style_id' in request.path_params:
        style_id = request.path_params['style_id']

    return await execute_from_starlette(
        maps_api.get_collection_map, request, collection_id, style_id
    )


async def get_processes(request: Request,
                        process_id: str | None = None) -> Response:
    """
    OGC API - Processes description endpoint

    :param request: Starlette Request instance
    :param process_id: identifier of process to describe

    :returns: Starlette HTTP Response
    """
    if 'process_id' in request.path_params:
        process_id = request.path_params['process_id']

    return await execute_from_starlette(processes_api.describe_processes,
                                        request, process_id)


async def get_jobs(request: Request,
                   job_id: str | None = None) -> Response:
    """
    OGC API - Processes jobs endpoint

    :param request: Starlette Request instance
    :param job_id: job identifier

    :returns: Starlette HTTP Response
    """

    if 'job_id' in request.path_params:
        job_id = request.path_params['job_id']

    if job_id is None:  # list of submit job
        return await execute_from_starlette(processes_api.get_jobs, request)
    else:  # get or delete job
        if request.method == 'DELETE':
            return await execute_from_starlette(processes_api.delete_job,
                                                request, job_id)
        else:  # Return status of a specific job
            return await execute_from_starlette(processes_api.get_jobs,
                                                request, job_id)


async def execute_process_jobs(request: Request,
                               process_id: str | None = None) -> Response:
    """
    OGC API - Processes jobs endpoint

    :param request: Starlette Request instance
    :param process_id: process identifier

    :returns: Starlette HTTP Response
    """

    if 'process_id' in request.path_params:
        process_id = request.path_params['process_id']

    return await execute_from_starlette(processes_api.execute_process,
                                        request, process_id)


async def get_job_result(request: Request,
                         job_id: str | None = None) -> Response:
    """
    OGC API - Processes job result endpoint

    :param request: Starlette Request instance
    :param job_id: job identifier

    :returns: HTTP response
    """

    if 'job_id' in request.path_params:
        job_id = request.path_params['job_id']

    return await execute_from_starlette(processes_api.get_job_result,
                                        request, job_id)


async def get_collection_edr_query(request: Request,
                                   collection_id: str | None = None,
                                   instance_id: str | None = None,
                                   location_id: str | None = None
                                   ) -> Response:
    """
    OGC EDR API endpoints

    :param collection_id: collection identifier
    :param instance_id: instance identifier
    :param location_id: location id of a /locations/<location_id> query

    :returns: HTTP response
    """

    if 'collection_id' in request.path_params:
        collection_id = request.path_params['collection_id']

    if 'instance_id' in request.path_params:
        instance_id = request.path_params['instance_id']

    if (request.url.path.endswith('instances') or
            (instance_id is not None and
             request.url.path.endswith(instance_id))):
        return await execute_from_starlette(
            edr_api.get_collection_edr_instances, request, collection_id,
            instance_id
        )

    if 'location_id' in request.path_params:
        location_id = request.path_params['location_id']
        query_type = 'locations'
    else:
        query_type = request['path'].split('/')[-1]

    return await execute_from_starlette(
        edr_api.get_collection_edr_query, request, collection_id,
        instance_id, query_type, location_id,
        skip_valid_check=True,
    )


async def collections(request: Request,
                      collection_id: str | None = None) -> Response:
    """
    OGC API collections endpoint

    :param request: Starlette Request instance
    :param collection_id: collection identifier

    :returns: Starlette HTTP Response
    """
    if 'collection_id' in request.path_params:
        collection_id = request.path_params['collection_id']

    return await execute_from_starlette(core_api.describe_collections, request,
                                        collection_id)


async def stac_catalog_root(request: Request) -> Response:
    """
    STAC root endpoint

    :param request: Starlette Request instance

    :returns: Starlette HTTP response
    """
    return await execute_from_starlette(stac_api.get_stac_root, request)


async def stac_catalog_path(request: Request) -> Response:
    """
    STAC endpoint

    :param request: Starlette Request instance

    :returns: Starlette HTTP response
    """
    path = request.path_params["path"]
    return await execute_from_starlette(stac_api.get_stac_path, request, path)


async def stac_landing_page(request: Request) -> Response:
    """
    STAC API landing page endpoint

    :param request: Starlette Request instance

    :returns: Starlette HTTP response
    """

    return await execute_from_starlette(stac_api.landing_page, request)


async def stac_search(request: Request) -> Response:
    """
    STAC API search endpoint

    :param request: Starlette Request instance

    :returns: Starlette HTTP response
    """

    return await execute_from_starlette(stac_api.search, request)


async def admin_config(request: Request) -> Response:
    """
    Admin endpoint

    :returns: Starlette HTTP Response
    """

    if request.method == 'GET':
        return await execute_from_starlette(admin_api.get_config_, request,
                                            alternative_api=ADMIN)
    elif request.method == 'PUT':
        return await execute_from_starlette(admin_api.put_config, request,
                                            alternative_api=ADMIN)
    elif request.method == 'PATCH':
        return await execute_from_starlette(admin_api.patch_config, request,
                                            alternative_api=ADMIN)


async def admin_config_resources(request: Request) -> Response:
    """
    Resources endpoint

    :returns: HTTP response
    """

    if request.method == 'GET':
        return await execute_from_starlette(admin_api.get_resources, request,
                                            alternative_api=ADMIN)
    elif request.method == 'POST':
        return await execute_from_starlette(admin_api.put_resource, request,
                                            alternative_api=ADMIN)


async def admin_config_resource(request: Request,
                                resource_id: str) -> Response:
    """
    Resource endpoint

    :param resource_id: resource identifier

    :returns: Starlette HTTP Response
    """

    if 'resource_id' in request.path_params:
        resource_id = request.path_params['resource_id']

    if request.method == 'GET':
        return await execute_from_starlette(admin_api.get_resource, request,
                                            resource_id,
                                            alternative_api=ADMIN)
    elif request.method == 'PUT':
        return await execute_from_starlette(admin_api.put_resource, request,
                                            resource_id,
                                            alternative_api=ADMIN)
    elif request.method == 'PATCH':
        return await execute_from_starlette(admin_api.patch_resource, request,
                                            resource_id,
                                            alternative_api=ADMIN)
    elif request.method == 'DELETE':
        return await execute_from_starlette(admin_api.delete_resource, request,
                                            resource_id,
                                            alternative_api=ADMIN)


class ApiRulesMiddleware:
    """ Custom middleware to properly deal with trailing slashes.
    See https://github.com/encode/starlette/issues/869.
    """

    def __init__(
            self,
            app: ASGIApp
    ) -> None:
        self.app = app
        self.prefix = API_RULES.get_url_prefix('starlette')

    async def __call__(self, scope: Scope,
                       receive: Receive, send: Send) -> None:
        if scope['type'] == "http" and API_RULES.strict_slashes:
            path = scope['path']
            if path == self.prefix:
                # If the root (landing page) is requested without a trailing
                # slash, redirect to landing page with trailing slash.
                # Starlette will otherwise throw a 404, as it does not like
                # empty Route paths.
                url = URL(scope=scope).replace(path=f"{path}/")
                response = RedirectResponse(url)
                await response(scope, receive, send)
                return
            elif path != f"{self.prefix}/" and path.endswith('/'):
                # Resource paths should NOT have trailing slashes
                response = Response(status_code=404)
                await response(scope, receive, send)
                return

        await self.app(scope, receive, send)


api_routes = [
    Route('/', landing_page),
    Route('/openapi', openapi),
    Route('/asyncapi', asyncapi),
    Route('/conformance', conformance),
    Route('/TileMatrixSets/{tileMatrixSetId}', get_tilematrix_set),
    Route('/TileMatrixSets', get_tilematrix_sets),
    Route('/collections/{collection_id:path}/schema', collection_schema),
    Route('/collections/{collection_id:path}/queryables', collection_queryables),  # noqa
    Route('/collections/{collection_id:path}/tiles', get_collection_tiles),
    Route('/collections/{collection_id:path}/tiles/{tileMatrixSetId}', get_collection_tiles_metadata),  # noqa
    Route('/collections/{collection_id:path}/tiles/{tileMatrixSetId}/metadata', get_collection_tiles_metadata),  # noqa
    Route('/collections/{collection_id:path}/tiles/{tileMatrixSetId}/{tile_matrix}/{tileRow}/{tileCol}', get_collection_items_tiles),  # noqa
    Route('/collections/{collection_id:path}/items', collection_items, methods=['GET', 'POST', 'OPTIONS']),  # noqa
    Route('/collections/{collection_id:path}/items/{item_id:path}', collection_items, methods=['GET', 'PUT', 'DELETE', 'OPTIONS']),  # noqa
    Route('/collections/{collection_id:path}/coverage', collection_coverage),  # noqa
    Route('/collections/{collection_id:path}/map', collection_map),
    Route('/collections/{collection_id:path}/styles/{style_id:path}/map', collection_map),  # noqa
    Route('/processes', get_processes),
    Route('/processes/{process_id}', get_processes),
    Route('/jobs', get_jobs),
    Route('/jobs/{job_id}', get_jobs, methods=['GET', 'DELETE']),
    Route('/processes/{process_id}/execution', execute_process_jobs, methods=['POST']),  # noqa
    Route('/jobs/{job_id}/results', get_job_result),
    Route('/collections/{collection_id:path}/position', get_collection_edr_query),  # noqa
    Route('/collections/{collection_id:path}/area', get_collection_edr_query),
    Route('/collections/{collection_id:path}/cube', get_collection_edr_query),
    Route('/collections/{collection_id:path}/radius', get_collection_edr_query),  # noqa
    Route('/collections/{collection_id:path}/trajectory', get_collection_edr_query),  # noqa
    Route('/collections/{collection_id:path}/corridor', get_collection_edr_query),  # noqa
    Route('/collections/{collection_id:path}/locations', get_collection_edr_query),  # noqa
    Route('/collections/{collection_id:path}/locations/{location_id}', get_collection_edr_query),  # noqa
    Route('/collections/{collection_id:path}/instances', get_collection_edr_query),  # noqa
    Route('/collections/{collection_id:path}/instances/{instance_id}', get_collection_edr_query),  # noqa
    Route('/collections/{collection_id:path}/instances/{instance_id}/position', get_collection_edr_query),  # noqa
    Route('/collections/{collection_id:path}/instances/{instance_id}/area', get_collection_edr_query),  # noqa
    Route('/collections/{collection_id:path}/instances/{instance_id}/cube', get_collection_edr_query),  # noqa
    Route('/collections/{collection_id:path}/instances/{instance_id}/radius', get_collection_edr_query),  # noqa
    Route('/collections/{collection_id:path}/instances/{instance_id}/trajectory', get_collection_edr_query),  # noqa
    Route('/collections/{collection_id:path}/instances/{instance_id}/corridor', get_collection_edr_query),  # noqa
    Route('/collections/{collection_id:path}/instances/{instance_id}/locations', get_collection_edr_query),  # noqa
    Route('/collections/{collection_id:path}/instances/{instance_id}/locations/{location_id}', get_collection_edr_query),  # noqa
    Route('/collections', collections),
    Route('/collections/{collection_id:path}', collections),
    Route('/stac', stac_catalog_root),
    Route('/stac/{path:path}', stac_catalog_path),
    Route('/stac-api', stac_landing_page),
    Route('/stac-api/search', stac_search, methods=['GET', 'POST'])
]

admin_routes = [
    Route('/admin/config', admin_config, methods=['GET', 'PUT', 'PATCH']),
    Route('/admin/config/resources', admin_config_resources, methods=['GET', 'POST']),  # noqa
    Route('/admin/config/resources/{resource_id:path}', admin_config_resource,
          methods=['GET', 'PUT', 'PATCH', 'DELETE'])
]

if CONFIG['server'].get('admin', False):
    ADMIN = Admin(CONFIG, OPENAPI)
    api_routes.extend(admin_routes)

url_prefix = API_RULES.get_url_prefix('starlette')
APP = Starlette(
    routes=[
        Mount(f'{url_prefix}/static', StaticFiles(directory=STATIC_DIR)),
        Mount(url_prefix or '/', routes=api_routes)
    ]
)

if url_prefix:
    # If a URL prefix is in effect, Flask allows the static resource URLs
    # to be written both with or without that prefix (200 in both cases).
    # Starlette does not allow this, so for consistency we'll add a static
    # mount here WITHOUT the URL prefix (due to router order).
    APP.mount(
        '/static', StaticFiles(directory=STATIC_DIR),
    )

# If API rules require strict slashes, do not redirect
if API_RULES.strict_slashes:
    APP.router.redirect_slashes = False
    APP.add_middleware(ApiRulesMiddleware)

# CORS: optionally enable from config.
if CONFIG['server'].get('cors', False):
    from starlette.middleware.cors import CORSMiddleware
    APP.add_middleware(
        CORSMiddleware,
        allow_origins=['*'],
        allow_methods=['*'],
        expose_headers=['*']
    )

try:
    OGC_SCHEMAS_LOCATION = Path(CONFIG['server']['ogc_schemas_location'])
except KeyError:
    OGC_SCHEMAS_LOCATION = None

if (OGC_SCHEMAS_LOCATION is not None and
        not OGC_SCHEMAS_LOCATION.name.startswith('http')):
    if not OGC_SCHEMAS_LOCATION.exists():
        raise RuntimeError('OGC schemas misconfigured')
    APP.mount(
        f'{url_prefix}/schemas', StaticFiles(directory=OGC_SCHEMAS_LOCATION)
    )


@click.command()
@click.pass_context
@click.option('--debug', '-d', default=False, is_flag=True, help='debug')
def serve(ctx, server=None, debug=False):
    """
    Serve pygeoapi via Starlette. Runs pygeoapi
    as a uvicorn server. Not recommend for production.

    :param server: `string` of server type
    :param debug: `bool` of whether to run in debug mode,
                    default log level is INFO

    :returns: void
    """

    log_level = 'info'
    if debug:
        log_level = 'debug'
    uvicorn.run(
        "pygeoapi.starlette_app:APP",
        reload=True,
        log_level=log_level,
        loop='asyncio',
        host=api_.config['server']['bind']['host'],
        port=api_.config['server']['bind']['port'])


if __name__ == "__main__":  # run locally, for testing
    serve()



# =================================================================
#
# Authors: Sander Schaminee <sander.schaminee@geocat.net>
#
# Copyright (c) 2021 GeoCat BV
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging
from typing import List, Union
from collections import OrderedDict
from copy import deepcopy

from babel import Locale
from babel import UnknownLocaleError as _UnknownLocaleError
from urllib import parse

LOGGER = logging.getLogger(__name__)

# Specifies the name of a request query parameter used to set a locale
QUERY_PARAM = 'lang'

# Cache Babel Locale lookups by string
_lc_cache = {}

# Cache translated configurations
_cfg_cache = {}


class LocaleError(Exception):
    """ General exception for any kind of locale parsing error. """
    pass


def str2locale(value: str, silent: bool = False) -> Union[Locale, None]:
    """
    Converts a web locale or language tag into a Babel Locale instance.

    .. note::   If `value` already is a Locale, it is returned as-is.

    :param value: A string containing a (web) locale (e.g. 'fr-CH')
                    or language tag (e.g. 'de').
    :param silent: If True (default = False), no errors will be raised
                   when parsing failed. Instead, `None` will be returned.

    :returns: babel.core.Locale or None

    :raises LocaleError:
    """

    if isinstance(value, Locale):
        return value

    loc = _lc_cache.get(value)
    if loc:
        # Value has been converted before: return cached Locale
        return loc

    try:
        loc = Locale.parse(value.strip().replace('-', '_'))
    except (ValueError, AttributeError) as err:
        LOGGER.warning(err)
        if not silent:
            raise LocaleError(f"invalid locale '{value}'")
    except _UnknownLocaleError as err:
        LOGGER.warning(err)
        if not silent:
            raise LocaleError(err)
    else:
        # Add to Locale cache
        _lc_cache[value] = loc

    return loc


def locale2str(value: Locale) -> str:
    """
    Converts a Babel Locale instance into a web locale string.

    :param value: babel.core.Locale

    :returns: A string containing a web locale (e.g. 'fr-CH')
              or language tag (e.g. 'de').

    :raises LocaleError:
    """

    if not isinstance(value, Locale):
        raise LocaleError(f"'{value}' is not of type {Locale.__name__}")
    return str(value).replace('_', '-')


def best_match(accept_languages, available_locales) -> Locale:
    """
    Takes an Accept-Languages sorted list (from header or request query params)
    and finds the best matching locale from a list of available locales.

    This function provides a framework-independent alternative to the
    `best_match()` function available in Flask/Werkzeug.

    If no match can be found for the Accept-Languages,
    the first available locale is returned.

    This function always returns a Babel Locale instance. If you require the
    web locale string, please use the :func:`locale2str` function.
    If you only ever need the language part of the locale, use the `language`
    property of the returned locale.

    .. note::   Any tag in the `accept_languages` string that is an invalid
                or unknown locale is ignored. However, if no
                `available_locales` are specified, a `LocaleError` is raised.

    :param accept_languages: A Locale or list of one or more languages.
                             This can be as simple as "de" for example,
                             but it's also possible to include a territory
                             (e.g. "en-US" or "fr_BE") or even a complex
                             list sorted by quality values, e.g.
                             ["fr-CH", "fr", "en", "de", "*"].
    :param available_locales: A list containing the available locales.
                              For example, a pygeoapi provider might only
                              support ["de", "en"].
                              Locales in the list can be specified as strings
                              (e.g. "nl-NL") or `Locale` instances.

    :returns: babel.core.Locale

    :raises LocaleError:
    """

    def get_match(locale_, available_locales_):
        """ Finds the first match of `locale_` in `available_locales_`. """
        if not locale_:
            return None
        territories_ = available_locales_.get(locale_.language, {})
        if locale_.territory in territories_:
            # Full match on language and territory
            return locale_
        if None in territories_:
            # Match on language only (generic, no territory)
            return Locale(locale_.language)
        if territories_:
            # Match on language but another territory (use first)
            return Locale(locale_.language, territory=territories_[0])
        # No match at all
        return None

    if not available_locales:
        raise LocaleError('No available locales specified')

    if isinstance(accept_languages, Locale):
        # If a Babel Locale was used as input, transform back into a string
        accept_languages = [locale2str(accept_languages)]

    if not isinstance(accept_languages, list):
        # If `accept_languages` is not a string, ignore it
        LOGGER.debug(f"ignoring invalid accept-languages '{accept_languages}'")
        accept_languages = []

    # Process supported locales
    prv_locales = OrderedDict()
    for a in available_locales:
        loc = str2locale(a)
        prv_locales.setdefault(loc.language, []).append(loc.territory)

    # Return best match from accepted languages
    for lang in accept_languages:
        loc = str2locale(lang, True)
        if not loc:
            LOGGER.debug(f"ignoring invalid accept-language '{lang}'")
            continue
        match = get_match(loc, prv_locales)
        if match:
            LOGGER.debug(f"'{match}' matches requested '{accept_languages}'")
            return match

    # Nothing matched: return the first available locale
    for lang, territories in prv_locales.items():
        match = Locale(lang, territory=territories[0])
        LOGGER.debug(f"No match found for language '{accept_languages}'; "
                     f"returning default locale '{match}'")
        return match


def translate(value: str, language: Union[Locale, str]):
    """
    If `value` is a language struct (where its keys are language codes
    and its values are translations for each language), this function tries to
    find and return the translation for the given `language`.

    If the given `value` is not a dict, the original value is returned.
    If the requested language does not exist in the struct,
    the first language value is returned. If there are no valid language keys
    in the struct, the original value is returned as well.

    If `language` is not a string or Locale, a LocaleError is raised.

    :param value: A value to translate. Typically either a string or
                  a language struct dictionary.
    :param language: A locale string (e.g. "en-US" or "en") or Babel Locale.

    :returns: A translated string or the original value.

    :raises LocaleError:
    """

    nested_dicts = isinstance(value, dict) and any(isinstance(v, dict)
                                                   for v in value.values())
    if not isinstance(value, dict) or nested_dicts:
        # Return non-dicts or dicts with nested dicts as-is
        return value

    # Validate language key by type (do not check if parsable)
    if not isinstance(language, (str, Locale)):
        raise LocaleError('language is not a str or Locale')

    # First try fast approach: directly fetch expected language key
    translation = value.get(locale2str(language)
                            if hasattr(language, 'language') else language)
    if translation:
        return translation

    # Find valid locale keys in language struct
    # Also maps Locale instances to actual key names
    loc_items = OrderedDict((str2locale(k, True), k) for k in value.keys())
    if not loc_items or None in loc_items:
        # Return as-is if not ALL keys in the struct are locales
        return value

    # Find best language match and return value by its key
    out_locale = best_match([language], loc_items.keys())
    return value[loc_items[out_locale]]


def translate_struct(struct: dict | List[dict],
                     locale_: Locale, is_config: bool = False):
    """
    Returns a copy of a given dict or list, where all language structs
    are filtered on the given locale, i.e. all language structs are replaced
    by translated values for the best matching locale.

    :param struct: A dict or list (of dicts) to filter/translate.
    :param locale_: The Babel Locale to filter on.
    :param is_config: If True, the struct is treated as a pygeoapi config.
                      This means that the first 2 levels won't be translated
                      and the translated struct is cached for speed.

    :returns: A translated dict or list
    """

    def _translate_dict(obj, level: int = 0):
        """ Recursive function to walk and translate a struct. """
        items = obj.items() if isinstance(obj, dict) else enumerate(obj)
        for k, v in items:
            if 0 <= level <= max_level and isinstance(v, (dict, list)):
                # Skip first 2 levels (don't translate)
                _translate_dict(v, level + 1)
                continue
            if isinstance(v, list):
                _translate_dict(v, level + 1)  # noqa
                continue
            tr = translate(v, locale_)
            if isinstance(tr, dict):
                # Look for language structs in next level
                _translate_dict(tr, level + 1)
            else:
                # Overwrite level with translated value
                obj[k] = tr

    max_level = 1 if is_config else -1
    result = {}
    if not struct:
        return result
    if not locale_:
        return struct

    # Check if we already translated the dict before
    result = _cfg_cache.get(locale_) if is_config else result
    if not result:
        # Create deep copy of config and translate/filter values
        result = deepcopy(struct)
        _translate_dict(result)

        # Cache translated pygeoapi configs for faster retrieval next time
        if is_config:
            _cfg_cache[locale_] = result

    return result


def set_response_language(headers: dict, *locale_: Locale):
    """
    Sets the Content-Language on the given HTTP response headers dict.

    :param headers: A dict of HTTP response headers.
    :param locale_: The Babel Locale(s) to which to set the
                    Content-Language header.
                    Multiple locales can be set for this header.
                    Note that duplicates will be removed.

    :raises LocaleError: if no valid Babel Locale was found.
    """

    if not hasattr(headers, '__setitem__'):
        LOGGER.warning(f"Cannot set headers on object '{headers}'")
        return

    locales = []
    for loc in locale_:
        try:
            loc_str = locale2str(loc)
        except LocaleError:
            if len(locale_) == 1:
                raise
        else:
            if loc_str not in locales:
                locales.append(loc_str)

    if not locales:
        raise LocaleError('no valid locales set')
    loc_str = ', '.join(locales)

    LOGGER.debug(f'Setting Content-Language to {loc_str}')
    headers['Content-Language'] = loc_str


def add_locale(url: str, locale_: str) -> str:
    """
    Adds a locale query parameter (e.g. 'lang=en-US') to a URL.
    If `locale_` is None or an empty string, the URL will be returned as-is.

    :param url: The web page URL (may contain query string).
    :param locale_: The web locale or language tag to append to the query.

    :returns: A new URL with a 'lang=<locale>' query parameter.

    :raises requests.exceptions.MissingSchema:
    """

    loc = str2locale(locale_, True)
    if not loc:
        # Validation of locale failed
        LOGGER.warning(
            f"Invalid locale '{locale_}': returning URL as-is")
        return url

    try:
        url_comp = parse.urlparse(url)
        params = dict(parse.parse_qsl(url_comp.query))
        params[QUERY_PARAM] = locale2str(loc)
        qstr = parse.urlencode(params, quote_via=parse.quote, safe='/')
        return parse.urlunparse((
            url_comp.scheme,
            url_comp.netloc,
            url_comp.path,
            url_comp.params,
            qstr,
            url_comp.fragment
        ))
    except (TypeError, ValueError):
        LOGGER.warning(
            f"Failed to append '{QUERY_PARAM}={loc}': returning URL as-is")  # noqa
    return url


def get_locales(config: dict) -> list:
    """
    Reads the configured locales/languages from the given configuration.
    The first Locale in the returned list should be the default locale.

    :param config: A pygeaapi configuration dict

    :returns: A list of supported Locale instances
    """

    srv_cfg = config.get('server', {})
    lang = srv_cfg.get('languages', srv_cfg.get('language', []))

    if isinstance(lang, str):
        LOGGER.info(f"pygeoapi only supports 1 language: {lang}")
        lang = [lang]
    if not isinstance(lang, list) or len(lang) == 0:
        LOGGER.error("Missing 'language(s)' key in config or bad value(s)")
        raise LocaleError('No languages have been configured')

    try:
        return [str2locale(loc) for loc in lang]
    except LocaleError as err:
        LOGGER.debug(err)
        raise LocaleError('Bad value in supported server language(s)')


def get_plugin_locale(
        config: dict,
        requested_locale: Union[str, None]) -> Union[Locale, None]:
    """
    Returns the supported locale (best match) for a plugin
    based on the requested raw locale string.
    Returns None if the plugin does not support any locales.
    Returns the default (= first) locale that the plugin supports
    if no match for the requested locale could be found.

    :param config: The plugin definition
    :param requested_locale: The requested locale string (or None)
    """

    plugin_name = f"{config.get('name', '')} plugin".strip()
    if not requested_locale:
        LOGGER.debug(f'No requested locale for {plugin_name}')
        requested_locale = ''

    LOGGER.debug(f'Requested {plugin_name} locale: {requested_locale}')
    locales = config.get('languages', config.get('language', []))
    if locales:
        if not isinstance(locales, list):
            locales = [locales]
        locale = best_match(requested_locale, locales)
        LOGGER.info(f'{plugin_name} locale set to {locale}')
        return locale

    LOGGER.info(f'{plugin_name} has no locale support')
    return None



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2026 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging

from paho.mqtt import client as mqtt_client

from pygeoapi.pubsub.base import BasePubSubClient, PubSubClientConnectionError

LOGGER = logging.getLogger(__name__)


class MQTTPubSubClient(BasePubSubClient):
    """MQTT client"""

    def __init__(self, publisher_def):
        """
        Initialize object

        :param publisher_def: provider definition

        :returns: pycsw.pubsub.mqtt.MQTTPubSubClient
        """

        super().__init__(publisher_def)
        self.type = 'mqtt'
        self.port = self.broker_url.port

        self.userdata = {}

        msg = f'Connecting to broker {self.broker_safe_url} with id {self.client_id}'  # noqa
        LOGGER.debug(msg)
        self.conn = mqtt_client.Client(mqtt_client.CallbackAPIVersion.VERSION2,
                                       client_id=self.client_id)

        self.conn.enable_logger(logger=LOGGER)

        if None not in [self.broker_url.username, self.broker_url.password]:
            LOGGER.debug('Setting credentials')
            self.conn.username_pw_set(
                self.broker_url.username,
                self.broker_url.password)

        if self.port is None:
            if self.broker_url.scheme == 'mqtts':
                self.port = 8883
            else:
                self.port = 1883

        if self.broker_url.scheme == 'mqtts':
            self.conn.tls_set(tls_version=2)

    def connect(self) -> None:
        """
        Connect to an MQTT broker

        :returns: None
        """

        try:
            self.conn.connect(self.broker_url.hostname, self.port)
            LOGGER.debug('Connected to broker')
        except Exception as err:
            raise PubSubClientConnectionError(err)

    def pub(self, channel: str, message: str, qos: int = 1) -> bool:
        """
        Publish a message to a broker/channel

        :param channel: `str` of channel
        :param message: `str` of message

        :returns: `bool` of publish result
        """

        LOGGER.debug(f'Publishing to broker {self.broker_safe_url}')
        LOGGER.debug(f'Channel: {channel}')
        LOGGER.debug(f'Message: {message}')

        result = self.conn.publish(channel, message, qos)
        LOGGER.debug(f'Result: {result}')

        # TODO: investigate implication
        # result.wait_for_publish()

        if result.is_published:
            LOGGER.debug('Message published')
            return True
        else:
            msg = f'Publishing error code: {result[1]}'
            LOGGER.warning(msg)
            return False

    def __repr__(self):
        return f'<MQTTPubSubClient> {self.broker_safe_url}'



# =================================================================
#
# Authors: Antonio Cerciello <anto.nio.cerciello@gmail.com>
#
# Copyright (c) 2022 Antonio Cerciello
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import pydantic
from pydantic import BaseModel
from typing import List, Optional


class VectorLayers(BaseModel):
    id: str
    description: Optional[str]
    minzoom: Optional[int]
    maxzoom: Optional[int]
    fields: Optional[dict]


class MVTTilesJson(BaseModel):
    tilejson: str = "3.0.0"
    name: Optional[str] = None
    tiles: Optional[str] = None
    minzoom: Optional[int] = None
    maxzoom: Optional[int] = None
    bounds: Optional[str] = None
    center: Optional[str] = None
    attribution: Optional[str] = None
    description: Optional[str] = None
    vector_layers: Optional[List[VectorLayers]] = None


if pydantic.VERSION.startswith('1'):
    def _dump(self, *, exclude_none: bool = False, **kwargs):
        return self.dict(exclude_none=exclude_none, **kwargs)

    MVTTilesJson.model_dump = _dump



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Francesco Martinelli <francesco.martinelli@ingv.it>
#
# Copyright (c) 2022 Tom Kralidis
# Copyright (c) 2024 Francesco Martinelli
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging

from pygeoapi.process.base import BaseProcessor, ProcessorExecuteError


LOGGER = logging.getLogger(__name__)

#: Process metadata and description
PROCESS_METADATA = {
    'version': '0.2.0',
    'id': 'hello-world',
    'title': {
        'en': 'Hello World',
        'fr': 'Bonjour le Monde'
    },
    'description': {
        'en': 'An example process that takes a name as input, and echoes '
              'it back as output. Intended to demonstrate a simple '
              'process with a single literal input.',
        'fr': 'Un exemple de processus qui prend un nom en entrée et le '
              'renvoie en sortie. Destiné à démontrer un processus '
              'simple avec une seule entrée littérale.',
    },
    'jobControlOptions': ['sync-execute', 'async-execute'],
    'keywords': ['hello world', 'example', 'echo'],
    'links': [{
        'type': 'text/html',
        'rel': 'about',
        'title': 'information',
        'href': 'https://example.org/process',
        'hreflang': 'en-US'
    }],
    'inputs': {
        'name': {
            'title': 'Name',
            'description': 'The name of the person or entity that you wish to'
                           'be echoed back as an output',
            'schema': {
                'type': 'string'
            },
            'minOccurs': 1,
            'maxOccurs': 1,
            'keywords': ['full name', 'personal']
        },
        'message': {
            'title': 'Message',
            'description': 'An optional message to echo as well',
            'schema': {
                'type': 'string'
            },
            'minOccurs': 0,
            'maxOccurs': 1,
            'keywords': ['message']
        }
    },
    'outputs': {
        'echo': {
            'title': 'Hello, world',
            'description': 'A "hello world" echo with the name and (optional)'
                           ' message submitted for processing',
            'schema': {
                'type': 'object',
                'contentMediaType': 'application/json'
            }
        }
    },
    'example': {
        'inputs': {
            'name': 'World',
            'message': 'An optional message.',
        }
    }
}


class HelloWorldProcessor(BaseProcessor):
    """Hello World Processor example"""

    def __init__(self, processor_def):
        """
        Initialize object

        :param processor_def: provider definition

        :returns: pygeoapi.process.hello_world.HelloWorldProcessor
        """

        super().__init__(processor_def, PROCESS_METADATA)
        self.supports_outputs = True

    def execute(self, data, outputs=None):
        mimetype = 'application/json'
        name = data.get('name')

        if name is None:
            raise ProcessorExecuteError('Cannot process without a name')

        message = data.get('message', '')
        value = f'Hello {name}! {message}'.strip()

        produced_outputs = {}
        if not bool(outputs) or 'echo' in outputs:
            produced_outputs = {
                'id': 'echo',
                'value': value
            }

        return mimetype, produced_outputs

    def __repr__(self):
        return f'<HelloWorldProcessor> {self.name}'
